import logging
import unittest
from typing import Callable

import numpy as np
from scipy import signal

from ...config.sweep_config import SweepConfig
from ..defaults.base_structures import Spectrogram
from ..math_relation import Relation
from ..math_signal import Signal
from ..math_sweep import Sweep
from ..math_uncalcsweep import ApriorUncalculatedSweep, UncalculatedSweep
from ..utility_functions.ftat_functions import dwell
from .test_relation import PreTestRelation
from .test_signal import PreTestSignal

log = logging.getLogger()
log.setLevel(logging.INFO)


def pre_get_spectrogramm() -> Callable[[np.ndarray, np.ndarray], Spectrogram]:
    def extract(x: np.ndarray, y: np.ndarray) -> Spectrogram:
        log.debug("In getter spectrogram")
        t = (x,)
        f = (y,)
        S = 1
        return Spectrogram(f=f, t=t, S=S)

    return extract


def get_ricker(points, a=4.0):
    return signal.ricker(points, a)


SweepConfig.spectrogram_method = pre_get_spectrogramm()


class TestUncalculatedSweep(unittest.TestCase):
    def test_input_data(self):

        dt = 0.001
        t = np.linspace(0.0, 10.0, int(10.0 / dt) + 1)

        a_t_array = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13]
        a_t_Relation = Relation(x=t, y=t)

        def a_t_func(t):
            return t * 2

        f_t_array = [1, 20, 30, 40, 50, 60, 70, 80, 90, 100]
        f_t_Relation = Relation(x=t, y=t * 10)

        def f_t_func(t):
            return t**2

        ts = np.linspace(0.0, 1.0, int(1.0 / dt) + 1)
        aprior_signal = Signal(ts, get_ricker(ts.size))
        aprior_spectr = aprior_signal.get_spectrum()

        abs_aprior_spectr = aprior_spectr.get_amp_spectrum()

        l_a_t = [a_t_array, a_t_Relation, a_t_func, None]
        l_f_t = [f_t_array, f_t_Relation, f_t_func, None]
        l_apr = [aprior_signal, aprior_spectr, abs_aprior_spectr]

        for t_k in [t]:
            for a_t_k in l_a_t:
                for f_t_k in l_f_t:
                    with self.subTest(t_k=t_k, a_t_k=a_t_k, f_t_k=f_t_k):

                        result = UncalculatedSweep(t=t_k, f_t=f_t_k, a_t=a_t_k)
                        self.assertIsInstance(result, UncalculatedSweep)

                        sweep = result(t_k)
                        self.assertIsInstance(sweep, Sweep)

                        sweep2 = result()
                        self.assertIsInstance(sweep2, Sweep)

        for t_k in [t]:
            for apr_k in l_apr:
                for k_start in [None, 0.25]:
                    for k_end in [None, 10.0]:
                        with self.subTest(
                            t_k=t_k, apr_k=apr_k, k_start=k_start, k_end=k_end
                        ):

                            result = ApriorUncalculatedSweep(
                                t=t_k,
                                aprior_data=apr_k,
                                ftat_method=dwell(k_start, k_end, 5),
                            )

                            self.assertIsInstance(result, ApriorUncalculatedSweep)

                            sweep = result(t_k)
                            self.assertIsInstance(sweep, Sweep)

                            sweep2 = result()
                            self.assertIsInstance(sweep2, Sweep)


class TestSweep(unittest.TestCase):

    pre_relation = PreTestRelation
    pre_signal = PreTestSignal

    def test_input(self):
        self.pre_relation.pre_test_input(self, Sweep)

    def test_math(self):
        self.pre_relation.pre_test_math(self, Sweep)
        self.pre_signal.pre_test_math(self, Sweep)

    def test_integrate_diff(self):
        self.pre_relation.pre_test_integrate_diff(self, Sweep)

    def test_interpolate_extrapolate(self):
        self.pre_relation.pre_test_interpolate_extrapolate(self, Sweep)

    def test_convolve_correlate(self):
        s1 = Sweep([1, 2, 3, 4, 5], [10, 20, 30, 40, 50])
        sp1 = s1.get_spectrum()
        s2 = Sweep([1, 2, 3, 4, 5], [10, 20, 30, 40, 50])
        sp2 = s2.get_spectrum()

        l = [s1, sp1, s2, sp2]

        for k in l:
            for k2 in l:
                self.subTest(k=k, k2=k2)
                self.pre_relation.pre_test_convolve_correlate(self, k, k2, Sweep)



"""The core of representation of project.

- - - 

The basic unit is the `sweep_design.math_signals.math_relation.Relation` class. 
It describes the relationship between two sequences as a function of one 
variable *y = f(x)*. It introduces basic math operations (-, +, *, /, 
exponentiation(**)), derivation, integration, 
correlation, convolution, and other operations.

- - -

The next two classes are `sweep_design.math_signals.math_signal.Signal` and 
`sweep_design.math_signals.math_signal.Spectrum`, which inherit from the 
`sweep_design.math_signals.math_relation.Relation` class, are covered 
and can be obtained from the other through a transformation. 
The `sweep_design.math_signals.math_signal.Signal` class provides a method 
to get the instance of `sweep_design.math_signals.math_signal.Spectrum` class 
(the forward Fourier transformation is used by default). 
In the `sweep_design.math_signals.math_signal.Spectrum` class, the method for 
obtaining an instance of the `sweep_design.math_signals.math_signal.Signal` class 
(the inverse Fourier transform is used by default). 

The following methods are defined in the 
`sweep_design.math_signals.math_signal.Spectrum` class: 

* phase subtraction 
* phase addition 
* obtaining an inverse filter 
* amplitude spectrum acquisition
* phase spectrum acquisition
* static method for obtaining the spectrum from the amplitude and phase spectra

Methods are presented in the `sweep_design.math_signals.math_signal.Signal` 
class:

* obtaining a phase spectrum  
* obtaining an amplitude spectrum  
* obtaining an inverse signal  
* adding a phase  
* subtracting a phase  

They also have the same methods from the inherited 
`sweep_design.math_signals.math_relation.Relation` class.

- - -

The following `sweep_design.math_signals.math_sweep.Sweep` class
inherits from the `sweep_design.math_signals.math_signal.Signal` class. 
It does not have additional methods, but only contains the fields of frequency 
versus time (**frequency_time**), the amplitude envelope (**amplitude_time**), 
the spectrogram (**spectrogram**) and the a priori signal (**aprior_signal**).

- - -

The `sweep_design.math_signals.math_sweep.Sweep` class is the result of 
the `sweep_design.math_signals.math_uncalcsweep.UncalculatedSweep` 
or `sweep_design.math_signals.math_uncalcsweep.ApriorUncalculatedSweep` 
classes as a result of calling their instances.

The constructor of the 
`sweep_design.math_signals.math_uncalcsweep.UncalculatedSweep` class receives 
as input time (**time**), a function of frequency change from time 
(**frequency_time**), and a function of amplitude envelope from time 
(**amplitude_time**). And when calling an instance of the 
`sweep_design.math_signals.math_uncalcsweep.UncalculatedSweep` class, 
you can get a sweep signal corresponding to the passed functions of changing 
the frequency from time and the amplitude envelope from time. 
Envelope functions with conform to time and frequency with conform to 
time can be represented as either sequences or functions (or callables).

The constructor of the 
`sweep_design.math_signals.math_uncalcsweep.ApriorUncalculatedSweep` class 
accepts **apriori_data** (an instance of the 
`sweep_design.math_signals.math_relation.Relation`, 
`sweep_design.math_signals.math_signal.Signal`, 
`sweep_design.math_signals.math_signal.Spectrum`, 
or `sweep_design.math_signals.math_sweep.Sweep`, which will be cast to a 
`sweep_design.math_signals.math_signal.Spectrum` instance) and a method to 
extract from the `sweep_design.math_signals.math_signal.Spectrum` 
instance functions of frequency versus time, amplitude envelope versus time, 
and a priori signal.

- - - 

The above classes use methods defined in configuration files and classes: 
`sweep_design.config.config.Config` class and
`sweep_design.config.sweep_config.SweepConfig`.
The execution of methods in the above Classes can be changed by changing the 
methods in the configuration files, either by changing them or by importing 
configuration classes and changing their methods.

The default methods to use are defined in 
`sweep_design.math_signals.defaults.methods` 
and `sweep_design.math_signals.defaults.sweep_methods`

- - -

`utility_functions` defines useful functions such as:

> * Function `sweep_design.math_signals.utility_functions.emd_analyse.get_IMFs_emd`
to analyse a signal using Empirical Mode Decomposition and contains instances that 
contain Intrinsic Mode Functions (IMFs).

> * Function `sweep_design.math_signals.utility_functions.emd_analyse.get_IMFs_ceemdan`
to analyse a signal using Complete Ensemble Empirical Mode Decomposition with 
Adaptive Noise and return `NamedSignal` instances that contain Intrinsic 
Mode Functions (IMFs).

> * `sweep_design.math_signals.utility_functions.time_axis.get_time`
function to create time axis descrided by np.ndarray retruning reault
equal np.linspace. 

> * `sweep_design.math_signals.utility_functions.tukey.tukey_a_t` 
function to build the envelope for sweep signal.

> * `sweep_design.math_signals.utility_functions.linear_f_t.f_t_linear_array` or 
`sweep_design.math_signals.utility_functions.linear_f_t.f_t_linear_function` 
to get the linear mean of the frequency-time function as function or array

> * `sweep_design.math_signals.utility_functions.ftat_functions.simple_freq2time`
function to get frequency versus time and amplitude 
versus time changes from an a priori spectrum to create a sweep signal. 
**In this implementation, the amplitude of change over time is a 
constant.** The spectrum of the resulting sweep signal will be equal to 
the prior spectrum.

> * `sweep_design.math_signals.utility_functions.ftat_functions.dwell` - 
function to get frequency versus time and amplitude versus time 
changes from an a priori spectrum to create a sweep signal. **In this 
implementation, the amplitude of the envelope monotonously increases 
in proportion to the change in frequency up to the cutoff frequency fc, 
after which the amplitude of the envelope is constant.** The spectrum 
of the resulting sweep signal will be equal to the prior spectrum.

> * `sweep_design.math_signals.utility_functions.sweep_correction.correct_sweep_without_window` - 
Using the EMD to extract the first IMF from the displacement.  

> * `sweep_design.math_signals.utility_functions.sweep_correction.correct_sweep` - 
Using the EMD to extract the first IMF from the displacement and apply a 
window in the star so that the displacement starts at zero.

> * `sweep_design.math_signals.utility_functions.sweep_correction_source.get_correction_for_source` - 
Sweep signal correction for realization on the vibration source.

For convenience, you can import functions in the following way
```python
from sweep_design.math_signals.utility_functions import get_time
```

- - -

Module `sweep_design.math_signals.prepared_sweeps` contains various ready-made sweep 
signals. To create them, you need to call the appropriate functions with 
the necessary parameters.

Functions are defined to create the following sweep signals:

* Linear sweep 
(`sweep_design.math_signals.prepared_sweeps.linear_sweep.get_linear_sweep`)
* Dwell sweep 
(`sweep_design.math_signals.prepared_sweeps.dwell_sweep.get_dwell_sweep`)
* Code Zinger 
(`sweep_design.math_signals.prepared_sweeps.code_zinger.get_code_zinger`)
* m - sequence 
(`sweep_design.math_signals.prepared_sweeps.m_sequence.get_m_sequence`)
* shuffle 
(`sweep_design.math_signals.prepared_sweeps.pseudorandom_shuffle.get_shuffle`)

For convenience, you can import functions in the following way
```python
from sweep_design.math_signals.prepared_sweeps import get_linear_sweep
```

"""

from .math_relation import Relation as Relation
from .math_signal import Spectrum as Spectrum
from .math_signal import Signal as Signal
from .math_sweep import Sweep as Sweep
from .math_uncalcsweep import UncalculatedSweep as UncalculatedSweep
from .math_uncalcsweep import ApriorUncalculatedSweep as ApriorUncalculatedSweep



from typing import Sequence, Union
from pathlib import Path

from loguru import logger

from ...math_signals import Relation
from ...named_signals import NamedRelation

from .default import get_path_file_name


def write_txt_file(
    relation: Union[Relation, NamedRelation],
    file_name: Union[str, Path] = None,
    header: Union[bool, Sequence[str]] = None,
) -> None:
    """Save the data to txt file."""
    file_name = get_path_file_name(
        relation, file_name, ".txt" if file_name is None else Path(file_name).suffix
    )

    x, y = relation.get_data()

    with file_name.open("w") as f:
        if header is None or header is True:
            f.write("x\ty\n")
        elif len(header) > 1:
            f.write(f"{header[0]}\t{header[1]}\n")

        f.writelines((f"{x}\t{y}\n" for x, y in zip(x, y)))

    logger.info("The file is saved in the following path: {}", file_name)



import numpy as np

from sweep_design.named_signals.named_sweep import NamedSpecrogram

from ..base_view.abc_common_framer_grapher.figure import CommonFigure, CommonFigureMaker
from ..base_view.abc_common_framer_grapher.frame import CommonFrameMaker

from ...math_signals import Relation
from ...named_signals import NamedRelation, NamedSweep

from .controller import InteracterPlotView

from typing import Any, Dict, Type, Union


class GeneralView:
    def __init__(
        self,
        grapher: Type[CommonFigureMaker],
        framer: Type[CommonFrameMaker[CommonFigure, InteracterPlotView]],
        figure_properties: Dict[str, Any],
        frame_properties: Dict[str, Any],
        checkbox_properties: Dict[str, Any],
        dropdown_properties: Dict[str, Any],
    ) -> None:

        self._grapher = grapher()
        self._controller = InteracterPlotView()

        figure = grapher.get_figure(figure_properties)
        self._frame = framer.get_figure_frame(figure, frame_properties)
        self._checkbox = framer.get_checkbox_frame(
            checkbox_properties, self._controller
        )
        self._dropdown = framer.get_dropdown_frame(
            dropdown_properties, self._controller
        )

        result_frame = self._frame.h_add(self._checkbox)
        result_frame = self._dropdown.v_add(result_frame)

        self._result_frame = result_frame

        self._controller.set_controlled_object(
            self._frame, self._checkbox, self._dropdown
        )

    @property
    def result_view(self):
        return self._result_frame

    def add_line(
        self,
        x: Union[NamedRelation, np.ndarray],
        y: np.ndarray = None,
        name: str = None,
        line_properties: Dict[str, Any] = None,
    ):

        if line_properties is None:
            line_properties = self._grapher.get_next_line_properties()

        if isinstance(x, (Relation, NamedRelation)):
            if name is None:
                name = str(x)
            line = self._frame.get_figure().add_line(
                *x.get_data(), name, None, line_properties
            )

        else:
            line = self._frame.get_figure().add_line(x, y, name, None, line_properties)

        self._checkbox.add_choise(
            name, None, {"color": line.get_color()}, self._controller
        )

    def add_image(
        self,
        x,
        y: np.ndarray = None,
        s_xy: np.ndarray = None,
        name: str = None,
        image_properties: Dict[str, Any] = None,
    ):

        if image_properties is None:
            image_properties = self._grapher.get_next_image_properties()

        if isinstance(x, (NamedSweep, NamedSpecrogram)):
            if name is None:
                name = str(x)
            if isinstance(x, NamedSweep):
                spectrogram = x.spectrogram
            else:
                spectrogram = x
            self._frame.get_figure().add_image(
                spectrogram.time,
                spectrogram.frequency,
                spectrogram.spectrogram_matrix,
                unique_name=name,
                groups=None,
                image_properties=image_properties,
            )

        else:
            self._frame.get_figure().add_image(x, y, s_xy, name, None, image_properties)

        self._dropdown.add_choise(name)

    def show(self):
        return self._result_frame.get_output()



"""
This module extends the `sweep_design.math_signals` module.

Repeats classes methods from the `sweep_design.math_signals` module and using 
composition of the math_signals classes. These methods have the same purposes.

Almost every repeating method has additional parameters: '**name**' 
and '**category**'.

How they will operate with each other is described by the 
`sweep_design.named_signals.header_signals` module.

The method used to create default names and categories is described in 
`sweep_design.named_signals.header_signals.defaults.names`

The configuration file `sweep_design.config.named_config` and 
`sweep_design.config.named_config.NamedConfig` contains important variables 
and methods that you can change.

You can also simply create a 
`sweep_design.named_signals.named_relation.NamedRelation` class from an 
instance of the `sweep_design.math_signals.math_relation.Relation` class by 
passing an instance of the `sweep_design.math_signals.math_relation.Relation` 
class to the constructor of the 
`sweep_design.named_signals.named_relation.NamedRelation` class. Also 
works for others.

The purpose of creating this module is to simplify the work with many 
examples of signals, spectra, sweeps and relations. Simplify their 
visualization, comparison and storage.

"""

from .named_relation import NamedRelation as NamedRelation
from .named_signal import NamedSignal as NamedSignal, NamedSpectrum
from .named_signal import NamedSpectrum as NamedSpectrum
from .named_sweep import NamedSweep as NamedSweep
from .named_uncalcsweep import NamedUncalcSweep as NamedUncalcSweep
from .named_uncalcsweep import NamedApriorUcalcSweep as NamedApriorUcalcSweep



"""Configuration file for frame and figure are used ipywidget and bokeh.

"""
from ..base_view.default_screen_properties import get_screen_params

result = get_screen_params()
width, height = result["width"], result["height"]


def get_figure_window(
    title: str,
    x_axis_label: str,
    y_axis_label: str,
    k_width=0.7,
    k_height=0.4,
) -> None:

    figure_window = {
        "width": width * k_width * 0.95,
        "height": height * k_height * 0.95,
        "title": title,
        "x_axis_label": x_axis_label,
        "y_axis_label": y_axis_label,
    }
    return figure_window


def get_figure_frame(k_width=0.7, k_height=0.4):

    figure_frame = {
        "width": f"{int(width*k_width)}px",
        "hieght": f"{int(height*k_height)}px",
        "border": "solid 1px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "location": "up",
    }
    return figure_frame


def get_checkbox_frame(description: str, no_data: str, k_width: float, k_height: float):

    checbox_frame = {
        "width": f"{int(width*k_width)}px",
        "height": f"{int(height*k_height)}px",
        "border": "solid 1px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "name_no_data": no_data,
        "description": description,
    }
    return checbox_frame


def get_dropdown_properties():

    dropdown_frame = {
        "border": "solid 0px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "name_no_data": "No data!",
        "description": "Data:",
    }

    return dropdown_frame


def get_default_line_properties():

    return



from typing import Dict, Union

from ..base_view.abc_common_framer_grapher.figure import CommonFigure
from ..base_view.abc_common_framer_grapher.frame import (
    CommonCheckBox,
    CommonCheckBoxFrame,
    CommonDropdownFrame,
    CommonFigureFrame,
)

from ..base_view.abc_common_framer_grapher.controller import CommonController


class InteracterPlotView(CommonController):
    def __init__(self) -> None:
        self._frame = None
        self._checkbox = None
        self._dropdown = None

    def set_controlled_object(
        self,
        frame: CommonFigureFrame[CommonFigure],
        checkbox: CommonCheckBoxFrame,
        dropdown: CommonDropdownFrame,
    ) -> None:

        self._frame = frame
        self._checkbox = checkbox
        self._dropdown = dropdown

    def widget_notify(self, obj: Union[CommonCheckBox, CommonDropdownFrame]):
        def listner(value: Dict):
            if isinstance(obj, CommonCheckBox):
                self._set_line_by_checkbox(obj, value["new"])

            if obj is self._dropdown:
                self._set_image(value["new"], True)
                self._set_image(value["old"], False)

        return listner

    def _set_line_by_checkbox(self, checkbox: CommonCheckBox, value: bool) -> None:

        line = self._frame.get_figure().get_line(checkbox.get_name())
        if line:
            line.set_visible(value)

    def _set_image(self, image_name: str, value: bool) -> None:

        image = self._frame.get_figure().get_image(image_name)
        if image:
            image.set_visible(value)



import numpy as np
from abc import ABCMeta, abstractmethod, abstractstaticmethod
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    Union,
    TypeVar,
    Generic,
)

Figure = TypeVar("Figure")
Line = TypeVar("Line")
Image = TypeVar("Image")


class CommonFigureMaker(metaclass=ABCMeta):
    @abstractstaticmethod
    def get_figure(figure_properties: Dict[str, Any]) -> "CommonFigure":
        pass

    @abstractmethod
    def get_next_line_properties(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def get_next_image_properties(self) -> Dict[str, Any]:
        pass


class CommonFigure(Generic[Figure], metaclass=ABCMeta):
    @abstractmethod
    def __init__(self, figure: Figure) -> None:
        self._figure = figure
        self._lines: Dict[str, CommonLine] = {}
        self._images: Dict[str, CommonImage] = {}

    @abstractmethod
    def add_extra_axis(self, axis_name: str, axis_titel: str, location: str) -> None:
        pass

    @abstractmethod
    def add_line(
        self,
        x: np.ndarray,
        y: np.ndarray,
        unique_name: str,
        groups: Union[str, Set[str]],
        line_properties: Dict[str, Any],
    ) -> "CommonLine":
        pass

    @abstractmethod
    def add_image(
        self,
        x: np.ndarray,
        y: np.ndarray,
        s_xy: np.ndarray,
        unique_name: str,
        groups: Set[str],
        image_properties: Dict[str, Any],
    ) -> None:
        pass

    @abstractmethod
    def add_infty_line(
        self,
        k: float,
        axis: Literal["x", "y"],
        unique_name: str,
        groups: Set[str],
        line_properties: Dict[str, Any],
    ) -> None:
        pass

    def get_figure(self) -> Figure:
        return self._figure

    def get_line(self, unique_name) -> Optional["CommonLine"]:
        return self._lines.get(unique_name, None)

    def get_image(self, unique_name) -> Optional["CommonImage"]:
        return self._images.get(unique_name, None)

    def get_lines_by_groups(self, groups: Union[str, Set[str]]) -> List["CommonLine"]:
        if isinstance(groups, str):
            groups = set([groups])
        return [
            line for line in self._lines.values() if groups.issubset(line.get_groups())
        ]

    def get_images_by_groups(self, groups: Union[str, Set[str]]) -> List["CommonImage"]:
        if isinstance(groups, str):
            groups = set([groups])
        return [
            image
            for image in self._images.values()
            if groups.issubset(image.get_groups())
        ]


class CommonGraphElement(metaclass=ABCMeta):
    def __init__(self, name: str, groups: Union[str, Set[str]]) -> None:
        self._name = name
        if isinstance(groups, str):
            groups = set([groups])
        self._groups = groups

    def set_groups(self, groups: Union[str, Iterable[str]]) -> None:
        if isinstance(groups, str):
            groups = set([groups])
        self._groups.update(groups)

    def get_groups(self) -> Set[str]:
        return self._groups

    def get_name(self) -> str:
        return self._name

    def remove_groups(self, groups: Union[str, Iterable[str]]) -> None:
        if isinstance(groups, str):
            groups = set([groups])
        for group in groups:
            self._groups.discard(group)

    @abstractmethod
    def set_visible(self, is_visible: bool) -> None:
        pass


class CommonLine(CommonGraphElement, Generic[Line]):
    def __init__(self, line: Line, name: str, groups: Union[str, Set[str]]) -> None:
        self._line = line
        super().__init__(name, groups)

    @abstractmethod
    def get_color(self) -> Any:
        pass


class CommonImage(CommonGraphElement, Generic[Image]):
    def __init__(self, image: Image, name: str, groups: Union[str, Set[str]]) -> None:
        self._image = image
        super().__init__(name, groups)



from ...named_signals.named_sweep import NamedSweep

from ..framemakers.ipywidgetframer.default_ipywidget_properties import (
    get_checkbox_frame,
    get_dropdown_frame,
    get_figure_frame,
)

from ..framemakers.ipywidgetframer.figwidget.bokehfig import (
    BokehFigureMaker,
    color_palette,
    default_image_properties,
    default_line_property,
)

from ..framemakers.ipywidgetframer.figwidget.default_bokeh_properties import (
    get_figure_window,
)

from ..framemakers.ipywidgetframer.ipywidgetframe import WidgetFrameMaker
from .pilor_bp_rm_view_builder import PilotRmBpViewBuilder


class PilotRMBPBokehView:
    def __init__(self, k_width=1.0, k_height=1.0) -> None:

        figure_frame_amplitude_phase_spectrum = get_figure_frame(
            k_width * 0.4, k_height * 0.3
        )
        figure_window_amp_spectrum = get_figure_window(
            "Amplitude spectrum",
            "Frequency, Hz",
            "Amlitude, N",
            k_width * 0.4,
            k_height * 0.3,
        )

        figure_window_phase_spectrum = get_figure_window(
            "Phase spectrum", "Frequency, Hz", "Phase", k_width * 0.4, k_height * 0.3
        )

        checkbox_signals_frame = get_checkbox_frame(
            "Signals:", "", k_width * 0.4, k_height * 0.3
        )
        checkbox_types_signals_frame = get_checkbox_frame(
            "Types signals:", "", k_width * 0.4, k_height * 0.3
        )

        figure_frame_main_sweep_spectrogram = get_figure_frame(
            k_width * 0.8, k_height * 0.4
        )
        figure_window_main_sweep = get_figure_window(
            "Sweep", "Time, s", "Force, N", k_width * 0.8, k_height * 0.3
        )

        figure_window_spectrogram = get_figure_window(
            "Spectrum", "Time, s", "Frequency, Hz", k_width * 0.8, k_height * 0.3
        )

        dropdown_spectrogram_frame = get_dropdown_frame(
            "Specrograms:", "No spectrogram!"
        )

        self._view_builder = PilotRmBpViewBuilder(
            BokehFigureMaker,
            WidgetFrameMaker,
            figure_frame_amplitude_phase_spectrum,
            figure_window_amp_spectrum,
            figure_frame_amplitude_phase_spectrum,
            figure_window_phase_spectrum,
            checkbox_signals_frame,
            checkbox_types_signals_frame,
            figure_frame_main_sweep_spectrogram,
            figure_window_main_sweep,
            dropdown_spectrogram_frame,
            figure_frame_main_sweep_spectrogram,
            figure_window_spectrogram,
        )

    def add_pilot_rm_bp(
        self, pilot: NamedSweep, reaction_mass: NamedSweep, base_plate: NamedSweep
    ) -> None:

        line_properties = default_line_property
        line_properties["line_color"] = next(color_palette)

        self._view_builder.add_choise_sweep(
            f"{str(pilot)}_pilot",
            f"{str(pilot)}_pilot",
            {"color": line_properties["line_color"]},
        )

        time_spectrogram, frequency_spectrogram = pilot.f_t.get_data()
        self._view_builder.add_line_spectrogram(
            time_spectrogram,
            frequency_spectrogram,
            str(pilot),
            set([str(pilot), f"{str(pilot)}_pilot"]),
            line_properties,
        )

        for signal in [pilot, reaction_mass, base_plate]:

            self._view_builder.add_choise_type_signal(
                str(signal),
                set([str(signal), f"{str(pilot)}_pilot"]),
                {"color": line_properties["line_color"]},
            )

            frequency, amplitude_spectrum = (
                signal.get_spectrum().get_amp_spectrum().get_data()
            )
            self._view_builder.add_amp_spectrum(
                frequency,
                amplitude_spectrum,
                str(signal),
                set([str(signal), f"{str(pilot)}_pilot"]),
                line_properties,
            )

            frequency_phase, phase_spectrum = (
                signal.get_spectrum().get_phase_spectrum().get_data()
            )
            self._view_builder.add_phase_sepctrum(
                frequency_phase,
                phase_spectrum,
                str(signal),
                set([str(signal), f"{str(pilot)}_pilot"]),
                line_properties,
            )

            time, amplitude = signal.get_data()
            self._view_builder.add_main_sweep_signals(
                time,
                amplitude,
                str(signal),
                set([str(signal), f"{str(pilot)}_pilot"]),
                line_properties,
            )

            spectrogram = signal.spectrogram
            time_spectrogram, frequency_spectrogram, spectrogram = (
                spectrogram.time,
                spectrogram.frequency,
                spectrogram.spectrogram_matrix,
            )

            self._view_builder.add_image_spectrogram(
                time_spectrogram,
                frequency_spectrogram,
                spectrogram,
                str(signal),
                set([str(signal), str(pilot), f"{str(pilot)}_pilot"]),
                default_image_properties,
            )
            self._view_builder.add_choise_dropdown_spectrogram(str(signal))
            line_properties["line_color"] = next(color_palette)

    def show(self):
        return self._view_builder.show()


def get_pilot_rm_bp_view_bokeh_ipywidget(k_width=1.0, k_height=1.0):
    return PilotRMBPBokehView(k_width, k_height)



from __future__ import annotations

from typing import Any, Dict, Set, Union

import ipywidgets as widgets

from ...base_view.abc_common_framer_grapher.controller import CommonController
from ...base_view.abc_common_framer_grapher.frame import (
    CommonCheckBox,
    CommonCheckBoxFrame,
    CommonController,
    CommonDropdownFrame,
    CommonFigureFrame,
    CommonFrame,
    CommonFrameMaker,
)


class WidgetFrameMaker(CommonFrameMaker):
    @staticmethod
    def get_figure_frame(figure, frame_properties: Dict[str, Any]) -> "FigureFrame":

        window = widgets.HBox([figure])
        window.layout = widgets.Layout(**frame_properties)
        return FigureFrame(figure, window)

    @staticmethod
    def get_checkbox_frame(
        frame_properties: Dict[str, Any], controller: CommonController
    ) -> "CheckBoxFrame":

        header_label = frame_properties["description"]
        label = widgets.HTML(value=f"<b><font color='black'>{header_label}:</b>")
        output = widgets.VBox([label])
        output.layout = widgets.Layout(**frame_properties)
        return CheckBoxFrame(output, controller)

    @staticmethod
    def get_dropdown_frame(
        dropdown_frame_properties: Dict[str, Any], controller: CommonController
    ) -> "DropdownFrame":
        w = widgets.Dropdown(
            options=[dropdown_frame_properties["name_no_data"]],
            value=dropdown_frame_properties["name_no_data"],
            description=dropdown_frame_properties["description"],
            disabled=False,
        )
        w.layout = widgets.Layout(**dropdown_frame_properties)
        return DropdownFrame(w, dropdown_frame_properties["name_no_data"], controller)


class WidgetFrame(CommonFrame[widgets.Box, CommonController]):
    def h_add(self, other: "WidgetFrame") -> "WidgetFrame":
        if not (self._output is None or other._output is None):
            return WidgetFrame(widgets.HBox([self._output, other._output]))
        elif self._output is None:
            return other
        else:
            return self

    def v_add(self, other: "WidgetFrame") -> "WidgetFrame":
        if not (self._output is None or other._output is None):
            return WidgetFrame(widgets.VBox([self._output, other._output]))
        elif self._output is None:
            return other
        else:
            return self

    def add_outputs(self, other: Any) -> None:
        self._output.children += (other,)

    def remove_outputs(self, other: Any) -> None:
        list_outputs = list(self._output.children)
        list_outputs.pop(list_outputs.index(other))
        self._output.children = tuple(list_outputs)


class FigureFrame(WidgetFrame, CommonFigureFrame):
    pass


class CheckBoxFrame(WidgetFrame, CommonCheckBoxFrame["CheckBox"]):
    def add_choise(
        self,
        choise: str,
        groups: str,
        checkbox_properties: Dict[str, Any],
        controller: CommonController,
    ) -> None:

        text = "____ {0}".format(choise)
        checkbox = widgets.Checkbox(
            value=True,
            description=f'<b><font color={checkbox_properties["color"]}>{text}</b>',
            disabled=False,
            indent=False,
        )
        choise_box = CheckBox(checkbox, choise, groups, controller)
        self.add_outputs(choise_box._output)
        self._checkboxs.update({choise: choise_box})

    def return_choise(self, choise: str) -> None:
        self._checkboxs[choise]._checkbox.value = True
        self.add_outputs(self._checkboxs[choise]._output)

    def remove_choise(self, choise: str) -> None:
        self._checkboxs[choise]._checkbox.value = False
        self.remove_outputs(self._checkboxs[choise]._output)


class DropdownFrame(WidgetFrame, CommonDropdownFrame[widgets.Dropdown]):
    def __init__(
        self,
        dropdown: widgets.Dropdown,
        no_data_name: str,
        controller: CommonController,
    ) -> None:
        super().__init__(dropdown, no_data_name, controller)
        self._dropdown.observe(self._controller.widget_notify(self), "value")

    def add_choise(self, choise: str) -> None:
        options = list(self._dropdown.options)
        options.append(choise)
        self._dropdown.options = options
        self._dropdown.value = choise

    def remove_choise(self, choise: str) -> None:
        options = list(self._dropdown.options)
        options.remove(choise)
        self._dropdown.options = options

    def get_choise(self) -> None:
        return self._dropdown.value


class CheckBox(WidgetFrame, CommonCheckBox[widgets.Checkbox, CommonController]):
    def __init__(
        self,
        output: widgets.Checkbox,
        name: str,
        groups: Union[str, Set[str], None],
        controller: CommonController,
    ) -> None:
        super().__init__(output, name, groups, controller)
        self._checkbox.observe(self._controller.widget_notify(self), "value")

    def set_select(self, value: bool) -> None:
        self._checkbox.value = value



from typing import Any, Callable, Optional, Tuple, Type, TypeVar, Union

import numpy as np

from ..config.named_config import NamedConfig
from ..math_signals.defaults.base_structures import MathOperation
from ..math_signals.math_relation import Relation
from .header_signals.base_header import HeaderBase
from .header_signals.relation_header import HeaderRelation

Num = Union[float, int, complex]
CR = TypeVar("CR", bound="NamedRelation")
CR2 = TypeVar("CR2", bound="NamedRelation")
InName = Union[HeaderBase, str, Callable[[], str]]
InData = Union[CR, Relation]
Any = Union[InData, Num]
T = TypeVar("T")

default_name_composed_relation = "DR"


class NamedRelation:

    extract_input_default = NamedConfig.extract_input

    def __init__(
        self, x: Any, y: Any = None, name: InName = None, category: str = None
    ) -> None:

        x, y = NamedConfig.extract_input(x, y)

        if category is None:
            category = "relation"

        self._relation = Relation(x, y)
        self.header = HeaderRelation(name, category)

    @property
    def x(self):
        return self._relation.x

    @property
    def y(self):
        return self._relation.y

    @property
    def dx(self):
        return self._relation.dx

    @property
    def relation(self):
        return self._relation

    @property
    def category(self):
        return self.header.category

    @classmethod
    def _convert_input(
        cls: Type[CR], data: Any, name: InName = None
    ) -> "NamedRelation":
        if not isinstance(data, NamedRelation):
            if not name:
                name = default_name_composed_relation
            return NamedRelation(cls.extract_input_default(data, None), name=name)
        return data

    def _operate(
        self: CR,
        other: Any,
        operation: MathOperation,
        name: Optional[InName],
        category: Optional[str],
    ) -> CR:

        if not isinstance(other, (NamedRelation, float, int, complex)):
            other = self._convert_input(other)

        relation = getattr(self._relation, operation.value)(other)
        header = getattr(self.header, operation.value)(other, name, category)
        return type(self)(relation, name=header)

    def get_data(self) -> Tuple[np.ndarray, np.ndarray]:
        return self._relation.get_data()

    def select_data(
        self: CR,
        x_start: Num = None,
        x_end: Num = None,
        name: InName = None,
        category: str = None,
    ) -> CR:

        if x_start is None:
            x_start = self._relation._x[0]

        if x_end is None:
            x_end = self._relation._x[-1]

        header = self.header.select_data(x_start, x_end, name, category)
        relation = self._relation.select_data(x_start, x_end)
        return type(self)(relation, name=header)

    def exp(self: CR, name: InName = None, category="relation") -> CR:
        header = self.header.exp(name, category)
        relation = self._relation.exp()
        return type(self)(relation, name=header)

    def diff(self: CR, name: InName = None, category="relation") -> CR:
        header = self.header.diff(name=name, category=category)
        relation = self._relation.diff()
        return type(self)(relation, name=header)

    def integrate(self: CR, name: InName = None, category="relation") -> CR:
        header = self.header.integrate(name, category)
        relation = self._relation.integrate()
        return type(self)(relation, name=header)

    def interpolate_extrapolate(
        self: CR, new_x, name: InName = None, category="relation"
    ) -> CR:
        header = self.header.interpolate_extrapolate(new_x, name, category)
        relation = self._relation.interpolate_extrapolate(new_x)
        return type(self)(relation, name=header)

    def shift(
        self: CR, x_shift: Num = 0.0, name: InName = None, category="relation"
    ) -> CR:
        header = self.header.shift(x_shift, name, category)
        relation = self._relation.shift(x_shift)
        return type(self)(relation, name=header)

    @classmethod
    def equalize(
        cls: Type[CR],
        cr1: Any,
        cr2: Any,
        name1: InName = None,
        name2: InName = None,
        category1: str = None,
        category2: str = None,
    ) -> Tuple[CR, CR2]:
        cr1 = cls._convert_input(cr1)
        cr2 = cls._convert_input(cr2)
        r1, r2 = Relation.equalize(cr1.relation, cr2.relation)
        header1, header2 = HeaderRelation.equalize(
            cr1.header, cr2.header, r1, name1, name2, category1, category2
        )
        return type(cr1)(r1, name=header1), type(cr2)(r2, name=header2)

    @classmethod
    def convolve(
        cls: Type[CR],
        cr1: Any,
        cr2: Any,
        name: InName = None,
        category: str = "relation",
    ) -> CR:
        cr1 = cls._convert_input(cr1)
        cr2 = cls._convert_input(cr2)
        relation = Relation.convolve(cr1.relation, cr2.relation)
        header = HeaderRelation.convolve(cr1.header, cr2.header, name, category)
        return cls(relation, name=header)

    @classmethod
    def correlate(
        cls: Type[CR],
        cr1: Any,
        cr2: Any,
        name: InName = None,
        category: str = "relation",
    ) -> CR:
        cr1 = cls._convert_input(cr1)
        cr2 = cls._convert_input(cr2)
        named = HeaderRelation.convolve(cr1.header, cr2.header, name, category)
        relation = Relation.correlate(cr1.relation, cr2.relation)
        return cls(relation, name=named)

    def __str__(self) -> str:
        return str(self.header)

    def __add__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.ADD, name, category)

    def __radd__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.RADD, name, category)

    def __sub__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.SUB, name, category)

    def __rsub__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.RSUB, name, category)

    def __mul__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.MUL, name, category)

    def __rmul__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.RMUL, name, category)

    def __truediv__(
        self: CR, other: Any, name: InName = None, category="relation"
    ) -> CR:
        return self._operate(other, MathOperation.TRUEDIV, name, category)

    def __rtruediv__(
        self: CR, other: Any, name: InName = None, category="relation"
    ) -> CR:
        return self._operate(other, MathOperation.RTRUEDIV, name, category)

    def __pow__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.POW, name, category)

    def __rpow__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self._operate(other, MathOperation.RPOW, name, category)

    def __iadd__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self.__add__(other, name, category)

    def __isub__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self.__sub__(other, name, category)

    def __imul__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self.__mul__(other, name, category)

    def __idiv__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self.__truediv__(other, name, category)

    def __ipow__(self: CR, other: Any, name: InName = None, category="relation") -> CR:
        return self.__pow__(other, name, category)

    def __getitem__(self, item: Union[float, slice]):
        if isinstance(item, float):
            x = self._relation._x 
            idx = (np.abs(x - item)).argmin()
            return x[idx], self._relation._y[idx]
        if isinstance(item, slice):
            return self.select_data(item.start, item.stop)


from typing import Any, Tuple

import numpy as np
from ...math_signals.defaults.base_structures import BadInputError, RelationProtocol


class ConverError(Exception):
    pass


x = np.ndarray
y = np.ndarray


def extract_input(x: Any, y: Any) -> Tuple[x, y]:
    """Simple data extraction from parameters."""
    if y is None:
        if isinstance(
            x,
            (
                tuple,
                list,
            ),
        ):
            out_x, out_y = np.array(x[0]), np.array(x[1])
        elif isinstance(x, dict):
            data = list(x.values())
            out_x, out_y = np.array(data[0]), np.array(data[1])
        elif isinstance(x, RelationProtocol):
            out_x, out_y = x.get_data()
        else:
            raise ConverError(f"Unable convert data: x - {x} and y - {y}")

    elif isinstance(y, np.ndarray) and isinstance(x, np.ndarray):
        out_x, out_y = x, y
    else:
        out_x, out_y = np.array(x), np.array(y)

    if out_x.size != out_y.size:
        raise BadInputError(
            f"Lengths of array x and array y is not eqaul.\n"
            f"Size of x is {out_x.size}. Size of y is {out_y.size}."
        )
    return out_x, out_y



from math import sin
from typing import Callable, Optional, TypeVar

import numpy as np
from loguru import logger

from ..math_relation import Relation
from .emd_analyse import get_IMFs_emd
from .tukey import tukey_a_t

Signal = TypeVar("Signal", bound=Relation)


def soft_clip(
    data: np.ndarray, limits: float, percent=0.85, coef: int = 1
) -> np.ndarray:
    """Custom function for correction."""
    hard_limit = limits
    linear_limit = limits * percent

    def select_choice(data: np.ndarray):

        amplitude = abs(data[0])

        if amplitude <= linear_limit:
            return data[1]

        if amplitude >= hard_limit:
            return ((hard_limit / amplitude) ** coef) * data[1]

        scale = hard_limit - linear_limit
        compression = scale * sin(np.pi / 2 * (amplitude - linear_limit) / scale)

        return (((linear_limit + compression) / amplitude) ** coef) * data[1]

    return np.apply_along_axis(select_choice, 1, data)


def get_correction_for_source(
    signal: Signal,
    reaction_mass: float = 1.0,
    limits: float = None,
    limits_persent=0.85,
    limit_iteration: Optional[int] = 10,
    window_percent=0.01,
    coef_function: Callable[[int], float] = lambda x: x,
) -> Signal:

    """Sweep signal correction for realization on the vibration source.

    Steps of corrections:
    1. Calculate displcament from force.
    2. Using EMD find first IMFs of displacement.
    3. Apply suppression amplitudy after limits.
    4. Apply window at the start to ensure a zero first amplitude.
    5. Return сalculated аorce from displacement

    Parametrs:
    > `signal`: `Relation` - signal to be corrected

    > `reaction_mass`: `float` = 1. - reaction mass of source

    > `limits`: `float` = None - displacement limitation of source

    > `limits_persent`: `float` = 0.85 from 0 to 1, determine the limits =
    `limits*limits_persent` up to which the displacement amplitude will
    not changed, after that, the limit amplitude will be changed using the
    `soft_clip` function.

    > `limit_iteration`: int = 10 - iterate corrections.

    > `window_percent`: float = 0.01 - apply window at the initial displacment
    to ensure a zero first amplitude.

    > `coef_function` - function to suppress.

    Returns:
    > `Relation` of Force.

    """
    displacement = signal.integrate().integrate() / reaction_mass
    imfs = get_IMFs_emd(displacement)

    d_array = np.vstack((imfs[0].y, signal.y[1:-1]))
    d_array = np.transpose(d_array)

    cnt = 0
    while True and limits is not None:
        cnt += 1

        logger.info("Iteration of correction: {}", cnt)

        result = soft_clip(d_array, limits, limits_persent, coef_function(cnt))

        force = type(signal)(signal.x[: result.size], result)

        new_displacement = force.integrate().integrate() / reaction_mass

        imfs = get_IMFs_emd(new_displacement)

        new_displacement = imfs[0]

        if np.all(np.abs(new_displacement.y) < limits):
            break

        if limit_iteration is not None and cnt + 1 > limit_iteration:
            break

        d_array = np.vstack((new_displacement.y, force.y[1:-1]))
        d_array = np.transpose(d_array)

    new_displacement = new_displacement.shift(new_displacement.dx * cnt)

    time = new_displacement.x
    window = Relation(time, tukey_a_t(time, time[-1] * window_percent, "left"))

    return (new_displacement * window).diff().diff() * reaction_mass



from typing import List
from ..math_relation import Relation


def get_code_zinger(
    segment_sweep: Relation, code_zinger: List[int] = [-1, -1, -1, 1], periods=1
):
    """Repeat the transmitted signal according to the Zniger code for n periods (n times)."""

    new_sweep = segment_sweep * code_zinger[0]

    for cnt, v in enumerate(code_zinger[1:], 1):
        new_sweep = new_sweep + v * segment_sweep.shift(
            cnt * segment_sweep.x[-1] + cnt * segment_sweep.dx
        )

    new_sweep_period = new_sweep

    for period in range(periods - 1):
        print(period)
        new_sweep_period += new_sweep.shift(
            (period + 1) * new_sweep_period.x[-1] + (period + 1) * new_sweep_period.dx
        )

    return new_sweep_period



from typing import Dict

from ..base_view.abc_common_framer_grapher.figure import CommonFigure
from ..base_view.abc_common_framer_grapher.frame import (
    CommonCheckBoxFrame,
    CommonDropdownFrame,
    CommonFigureFrame,
)


class InteracterPilotRmBp:
    def __init__(
        self,
        dropdown_spectrogram_frame: CommonDropdownFrame = None,
        signal_figure_frame: CommonFigureFrame[CommonFigure] = None,
        amp_spectrum_figure_frame: CommonFigureFrame[CommonFigure] = None,
        phase_spectrum_figure_frame: CommonFigureFrame[CommonFigure] = None,
        spectogramma_figura_frame: CommonFigureFrame[CommonFigure] = None,
        checkbox_signal_frame: CommonCheckBoxFrame[CommonFigure] = None,
        checkbox_type_signal_frame: CommonCheckBoxFrame[CommonFigure] = None,
    ) -> None:
        self._dropdown_spectrogram_frame = dropdown_spectrogram_frame
        self._signal_figure_frame = signal_figure_frame
        self._amp_spectrum_figure_frame = amp_spectrum_figure_frame
        self._phase_spectrum_figure_frame = phase_spectrum_figure_frame
        self._spectogram_figura_frame = spectogramma_figura_frame
        self._checkbox_signals_frame = checkbox_signal_frame
        self._checkbox_types_signals_frame = checkbox_type_signal_frame

    def set_controled_objects(
        self,
        dropdown_spectrogram_frame: CommonDropdownFrame,
        signal_figure_frame: CommonFigureFrame,
        amp_spectrum_figure_frame: CommonFigureFrame[CommonFigure],
        phase_spectrum_figure_frame: CommonFigureFrame[CommonFigure],
        spectogramma_figura_frame: CommonFigureFrame[CommonFigure],
        checkbox_signal_frame: CommonCheckBoxFrame[CommonFigure],
        checkbox_type_signal_frame: CommonCheckBoxFrame[CommonFigure],
    ) -> None:
        self._dropdown_spectrogram_frame = dropdown_spectrogram_frame
        self._signal_figure_frame = signal_figure_frame
        self._amp_spectrum_figure_frame = amp_spectrum_figure_frame
        self._phase_spectrum_figure_frame = phase_spectrum_figure_frame
        self._spectogram_figura_frame = spectogramma_figura_frame
        self._checkbox_signals_frame = checkbox_signal_frame
        self._checkbox_types_signals_frame = checkbox_type_signal_frame

    def widget_notify(self, obj):
        def listner(value: Dict):
            if obj in self._checkbox_signals_frame.get_checkboxs().values():
                self._call_checkbox_signal(obj.get_name(), value["new"])

            if obj in self._checkbox_types_signals_frame.get_checkboxs().values():
                self._call_checkbox_type_signal(obj.get_groups(), value["new"])

            if obj is self._dropdown_spectrogram_frame:
                self._call_dropdown_spectrogram_frame(value["new"], True)
                self._call_dropdown_spectrogram_frame(value["old"], False)

        return listner

    def _call_checkbox_signal(self, checkbox_name: str, value: bool) -> None:
        for lines in self._amp_spectrum_figure_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for lines in self._phase_spectrum_figure_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for lines in self._signal_figure_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for lines in self._spectogram_figura_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for images in self._spectogram_figura_frame.get_figure().get_images_by_groups(
            checkbox_name
        ):
            images.set_visible(value)

            if value:
                self._dropdown_spectrogram_frame.add_choise(images.get_name())
            else:
                self._dropdown_spectrogram_frame.remove_choise(images.get_name())

        name_checkboxes = []
        for checkbox in self._checkbox_types_signals_frame.get_checkboxs().values():
            if checkbox_name in checkbox.get_groups():
                name_checkboxes.append(checkbox.get_name())

        for name in name_checkboxes:
            if value:
                self._checkbox_types_signals_frame.return_choise(name)
            else:
                self._checkbox_types_signals_frame.remove_choise(name)

    def _call_checkbox_type_signal(self, checkbox_name: str, value: str) -> None:

        for lines in self._amp_spectrum_figure_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for lines in self._phase_spectrum_figure_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for lines in self._signal_figure_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

        for lines in self._spectogram_figura_frame.get_figure().get_lines_by_groups(
            checkbox_name
        ):
            lines.set_visible(value)

    def _call_dropdown_spectrogram_frame(self, image_name: str, value: bool) -> None:
        image = self._spectogram_figura_frame.get_figure().get_image(image_name)
        if image:
            image.set_visible(value)



import logging
from math import sqrt
from typing import Any, Callable, Tuple, Union

import numpy as np
from numpy.typing import NDArray

from ..config.sweep_config import SweepConfig

from .defaults import sweep_methods as dfsm
from .defaults.base_structures import BadInputError
from .defaults.sweep_methods import Ftatr
from .math_relation import Relation
from .math_signal import Spectrum
from .math_sweep import Sweep

CallFtatMethod = Callable[[Spectrum], Tuple[np.ndarray, np.ndarray, np.ndarray]]


class UncalculatedSweep:
    """The `UncalculatedSweep` class prepares for the calculation of the signal sweep (`Sweep`).

    The get_info_from_ftat function is used to extract the frequency versus
    `time` and amplitude versus time functions from the passed `frequency_time`
    and `amplitude_time` parameters.

    **Properties**:

    > **time**: `Union[RelationProtocol, NDArray]` = `None`
    The Relation class, or a class derived from the Relation class, or
    an array_like object containing numbers(real or complex).

    > **frequency_time**: `Ftatr` = `None`
    This parameter, which describes changes in frequency over time,
    can be either an array_like, or an object from which an instance
    of the `Relation` class will be created, or an instance of the
    `Relation` class, or a callable object that returns a numeric sequence.
    If `None`, then the linear function f = t will be used.

    > **amplitude_time**: `Ftatr` = `None`
    This parameter, which describes changes in amplitude modulation
    over time, can be either an array_like, or an object from which
    an instance of the `Relation` class will be created, or an instance
    of the `Relation` class, or a callable object that
    returns a numeric sequence.
    If `None`, then the function will be assumed to be constant
    and equal to 1.

    """

    def __init__(
        self,
        time: NDArray = None,
        frequency_time: Union[Ftatr, NDArray] = None,
        amplitude_time: Union[Ftatr, NDArray] = None,
    ) -> None:

        self._integrate_function_default = SweepConfig.integrate_function

        time = time if time is None else np.array(time)
        if not (
            isinstance(frequency_time, (np.ndarray, Relation))
            or callable(frequency_time)
            or frequency_time is None
        ):
            frequency_time = np.array(frequency_time)

        if not (
            isinstance(amplitude_time, (np.ndarray, Relation))
            or callable(amplitude_time)
            or amplitude_time is None
        ):
            amplitude_time = np.array(amplitude_time)

        (
            self._time,
            self._frequency_time,
            self._amplitude_time,
        ) = dfsm.get_info_from_ftat(time, frequency_time, amplitude_time)

    def __call__(self, time: NDArray = None, tht0=0.0) -> Sweep:
        """Calling an instance of the class to calculate the sweep signal.

        Properties:

        > **time**: `NDArray` = `None`
        The number sequence determines the time.

        > **tht0**: `float` = 0.
        Zero phase.

        Return
        > `Sweep`
        Returns an instance of the Sweep class - the calculated sweep signal.

        If time is not passed or equals None, then the time sequence created
        when the class instance was initialized will be used.

        """
        logging.info(
            "Calling uncalculated sweep.\n"
            "with params:\nfrequency_time={0}\namplitude_time={1}\ntime={2}"
            "".format(self._frequency_time, self._amplitude_time, time)
        )

        if time is None and self._time is None:
            raise BadInputError("Not enough data: time")

        elif time is None and self._time is not None:
            calc_time = self._time
        elif time is not None:
            calc_time = time

        if self._frequency_time.__name__ == "interpolate_time":
            tht = self._array_tht(self._frequency_time(calc_time))
        else:
            tht = self._func_tht(self._frequency_time)

        sweep = self._amplitude_time(calc_time) * np.sin(tht(calc_time) + tht0)

        amplitude_time = Relation(calc_time, self._amplitude_time(calc_time))
        frequency_time = Relation(calc_time, self._frequency_time(calc_time))

        return Sweep(
            time=calc_time,
            amplitude=sweep,
            frequency_time=frequency_time,
            amplitude_time=amplitude_time,
        )

    def _func_tht(
        self, frequency_time: Callable[[np.ndarray], np.ndarray]
    ) -> Callable[[np.ndarray], np.ndarray]:
        """Functional representation of angular sweep."""

        def result(time: np.ndarray) -> np.ndarray:
            return self._integrate_function_default(frequency_time, time)

        return result

    def _array_tht(
        self, frequency_time: np.ndarray
    ) -> Callable[[np.ndarray], np.ndarray]:
        """Angular sweep represented by a numerical sequence."""

        def result(time: np.ndarray) -> np.ndarray:
            frequency_time_relation = Relation(time, frequency_time)
            return np.append([0.0], 2 * np.pi * frequency_time_relation.integrate().y)

        return result


class ApriorUncalculatedSweep(UncalculatedSweep):
    """`ApriorUncalculatedSweep`

    Class for constructing a sweep signal from a priori data (from another
    signal or spectrum).

    The calculation of the change in frequency with time (`frequency_time`)
    and the amplitude envelope with time (`amplitude_time`) will depend on
    the a priori data (`aprior_data`) and on the method (`ftat_method`)
    by which they will be calculated.

    If the conversion method (`ftat_method`) is not defined or `None`,
    then the method is taken from the `SweepConfig` class `freq2time`.
    Method can be overridden if necessary
    (`sweep_design.math_signals.config.SweepConfig.freq2time`).
    The extracted frequency over time (`frequency_time`) and the amplitude
    envelope over time (`amplitude_time`) will be send to the
    `UncalculatedSweep` constructor.
    """

    def __init__(
        self,
        time: Any = None,
        aprior_data: Any = None,
        ftat_method: CallFtatMethod = None,
    ) -> None:

        if not ftat_method:
            ftat_method = SweepConfig.freq2time

        (
            frequency_time,
            amplitude_time,
            self._aprior_signal,
        ) = dfsm.get_info_from_aprior_data(time, aprior_data, ftat_method)
        super().__init__(time, frequency_time, amplitude_time)

    def __call__(self, time: Any = None, tht0=0.0, is_normolize=True) -> Sweep:
        """Calculate the sweep and normolise it."""
        sweep = super().__call__(time=time, tht0=tht0)

        if is_normolize:
            norm_sweep = sweep.get_norm()
            norm_aprior = self._aprior_signal.get_norm()
            norm = sqrt(norm_aprior) / sqrt(norm_sweep)
            sweep.amplitude_time *= norm
            sweep *= norm

        sweep.aprior_signal = self._aprior_signal
        return sweep



import copy
import itertools
from typing import Any, Callable, Dict, Literal, Set, Union

import ipywidgets as widgets
import numpy as np
from bokeh.io import output_notebook, push_notebook, show
from bokeh.models import DataRange1d, LinearAxis, Span
from bokeh.models.renderers import GlyphRenderer
from bokeh.palettes import Set1_9, Set2_8
from bokeh.plotting import Figure, figure
from loguru import logger

from ....base_view.abc_common_framer_grapher.figure import (
    CommonFigure,
    CommonFigureMaker,
    CommonImage,
    CommonLine,
)

output_notebook()

_colors = list(Set1_9)
_colors.extend(list(Set2_8))
color_palette = itertools.cycle(_colors)

default_line_property = {"line_color": None, "line_width": 2, "line_dash": "solid"}

default_image_properties = {"palette": "Viridis256", "level": "image"}


class BokehFigureMaker(CommonFigureMaker):
    def __init__(
        self, default_line_property=default_line_property, color_palette=color_palette
    ) -> None:
        self.default_line_property = default_line_property
        self.color_palette = color_palette
        self.default_image_propertis = default_image_properties

    @staticmethod
    def get_figure(figure_properties: Dict[str, Any]) -> "BokehFigure":

        fig = figure(
            title=figure_properties["title"],
            title_location="above",
            plot_width=int(figure_properties["width"]),
            plot_height=int(figure_properties["height"]),
        )
        fig.xaxis.axis_label = figure_properties["x_axis_label"]
        fig.yaxis.axis_label = figure_properties["y_axis_label"]
        return BokehFigure(fig)

    def get_next_line_properties(self) -> Dict[str, Any]:
        line_properties = copy.deepcopy(self.default_line_property)
        line_properties["line_color"] = next(self.color_palette)
        return line_properties

    def get_next_image_properties(self) -> Dict[str, Any]:
        return {"palette": "Spectral11"}


class MetaFigure(type(widgets.Output), type(CommonFigure)):
    pass


class BokehFigure(widgets.Output, CommonFigure[Figure], metaclass=MetaFigure):
    def __init__(self, figure: Figure) -> None:
        super().__init__()
        self._figure = figure
        self._handle = None
        self._lines = {}
        self._images = {}
        self.on_displayed(lambda x: x.set_handle())

    def set_handle(self):
        self.clear_output()
        with self:
            self._handle = show(self._figure, notebook_handle=True)

    def get_handle(self):
        return self._handle

    def update(self):
        push_notebook(handle=self._handle)

    def add_extra_axis(self, axis_name: str, axis_label: str, location="right") -> None:
        self._figure.extra_y_ranges[axis_name] = DataRange1d()
        self._figure.add_layout(
            LinearAxis(y_range_name=axis_name, axis_label=axis_label), location
        )

    def get_axis(self, axis_name: str = None, type_axis: str = "y"):
        if type_axis == "x":
            if axis_name in self._figure.extra_x_ranges:
                return self._figure.extra_x_ranges[axis_name]
            else:
                return self._figure.x_range
        elif type_axis == "y":
            if axis_name in self._figure.extra_y_ranges:
                return self._figure.extra_y_ranges[axis_name]
            else:
                return self._figure.y_range

    def set_limits(
        self,
        range_axis: DataRange1d,
        min_z=0,
        max_z=1,
    ):

        range_axis.start = min_z + min_z * 0.25
        range_axis.end = max_z + max_z * 0.25

    def add_line(
        self,
        x: np.ndarray,
        y: np.ndarray,
        name: str,
        groups: Set[str],
        line_property: Dict[str, Any],
    ) -> "BokehLine":

        line = self._figure.line(x, y, **line_property)
        bokeh_line = BokehLine(line, name, groups, self.update)
        self._lines.update({name: bokeh_line})
        return bokeh_line

    def add_image(
        self,
        x: np.ndarray,
        y: np.ndarray,
        s_xy: np.ndarray,
        unique_name: str,
        groups: Union[str, Set[str]],
        image_properties: Dict[str, Any],
    ) -> None:

        if isinstance(groups, str):
            groups = set([groups])

        image = self._figure.image(
            image=[s_xy],
            x=x[0],
            y=y[0],
            dw=x[-1] - x[0],
            dh=y[-1] - y[0],
            **image_properties
        )
        image.level = "underlay"
        self._images.update(
            {unique_name: BokehImage(image, unique_name, groups, self.update)}
        )

    def add_infty_line(
        self,
        k: np.ndarray,
        axis: Literal["x", "y"],
        unique_name: str,
        groups: Set[str],
        line_properties: Dict[str, Any],
    ) -> None:

        if axis == "x":
            line = Span(location=k, dimension="height", **line_properties)
        elif axis == "y":
            line = Span(location=k, dimension="width", **line_properties)
        else:
            line = None

        self._figure.add_layout(line)
        bokeh_line = BokehLine(line, unique_name, groups, self.update)
        self._lines.update({unique_name: bokeh_line})


class BokehLine(CommonLine[GlyphRenderer]):
    def __init__(
        self, line: GlyphRenderer, name: str, groups: Union[str, Set[str]], update: Any
    ) -> None:
        super().__init__(line, name, groups)
        self._update = update

    def get_color(self) -> str:
        return list(self._line.glyph.references())[0].line_color

    def set_visible(self, check: bool) -> None:
        self._line.visible = check
        self._update()


class BokehImage(CommonImage[GlyphRenderer]):
    def __init__(
        self,
        image: GlyphRenderer,
        name: str,
        groups: Set[str],
        update: Callable[..., None],
    ) -> None:
        super().__init__(image, name, groups)
        self._update = update

    def set_visible(self, is_visible: bool) -> None:
        self._image.visible = is_visible
        self._update()



from ..base_view.default_screen_properties import get_screen_params

result = get_screen_params()

width, height = result["width"], result["height"]

layout_categoies = ["source"]


def get_figure_window(k_width=1.0, k_height=1.0):

    figure_window = {
        "width": width * k_width * 0.95,
        "height": height * k_height * 0.95,
        "title": "Signals",
        "x_axis_label": "Time, s",
        "y_axis_label": "Force, N",
        "extra_range_y_name": "displacement",
        "y_axis_label_2": "Position, m",
    }

    return figure_window


def get_figure_frame(k_width=1.0, k_height=1):

    figure_frame = {
        "width": f"{int(width*k_width)}px",
        "hieght": f"{int(height*k_height)}px",
        "border": "solid 1px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "location": "up",
    }

    return figure_frame


def get_checkbox_frame(k_width=1.0, k_height=1.0):

    checbox_frame = {
        "width": f"{int(width*0.1)}px",
        "height": f"{int(height*0.4)}px",
        "border": "solid 1px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "name_no_data": "No sweep signals!",
        "description": "Sweep signals",
    }

    return checbox_frame


def get_dropdown_frame():
    dropdown_frame = {
        "border": "solid 0px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "name_no_data": "No sweep signals!",
        "description": "Sweep signals:",
    }

    return dropdown_frame



from typing import Literal
import numpy as np
from scipy.signal.windows import tukey


def tukey_a_t2(time: np.ndarray, t_tapper: float) -> np.ndarray:
    """Tukey function to build the envelop for sweep signal.

    Parametrs:
    > time: np.ndarray - time axis
    > t_tapper: float - shape of tukey window in time.
    Returns:
    > envelope: np.ndarray
    """

    if t_tapper <= time[int(time.size / 2)]:
        tapper = time[time <= t_tapper].size * 2 / time.size
    else:
        tapper = 1.0
    return tukey(time.size, alpha=tapper)


def tukey_a_t(
    time: np.ndarray,
    t_tapper: float,
    location: Literal["left", "right", "both"] = "both",
):

    if t_tapper <= time[int(time.size / 2)]:
        tapper = time[time <= t_tapper].size * 2 / time.size
    else:
        tapper = 1.0

    result = tukey(time.size, alpha=tapper)

    if location == "both":
        return result

    result = np.append(
        result[: int(time.size / 2)], np.ones(time.size - int(time.size / 2))
    )

    if location == "left":
        return result

    if location == "right":
        return np.flip(result)

    return np.ones(time.size)



from typing import Any, Dict, Type

from ...named_signals.named_sweep import NamedSweep
from ..base_view.abc_common_framer_grapher.figure import CommonFigure, CommonFigureMaker
from ..base_view.abc_common_framer_grapher.frame import CommonFrameMaker
from .controller import InteracterSource


class ViewSource:
    def __init__(
        self,
        grapher: Type[CommonFigureMaker],
        framer: Type[CommonFrameMaker[CommonFigure, InteracterSource]],
        properties_source_figure_window: Dict[str, Any],
        properties_source_figure_frame: Dict[str, Any],
        properties_dropdown_signals_frame: Dict[str, Any],
        properties_checkbox_types_signals_frame: Dict[str, Any],
        line_properties_sweep,
        line_properties_envelope,
        line_properties_displacement,
        line_properties_limits,
        reaction_mass=1,
        limits=None,
    ) -> None:
        self._line_properties_sweep = line_properties_sweep
        self._line_properties_envelope = line_properties_envelope
        self._line_properties_displacement = line_properties_displacement
        self._line_properties_limits = line_properties_limits
        controller = InteracterSource()

        source_figure = grapher.get_figure(properties_source_figure_window)
        source_figure.add_extra_axis(
            properties_source_figure_window["extra_range_y_name"],
            properties_source_figure_window["y_axis_label_2"],
        )
        self._source_figure_frame = framer.get_figure_frame(
            source_figure, properties_source_figure_frame
        )
        self._dropdown_signals_frame = framer.get_dropdown_frame(
            properties_dropdown_signals_frame, controller
        )
        self._checkbox_types_signals_frame = framer.get_checkbox_frame(
            properties_checkbox_types_signals_frame, controller
        )

        result_frame = self._source_figure_frame.h_add(
            self._checkbox_types_signals_frame
        )
        result_frame = self._dropdown_signals_frame.v_add(result_frame)

        self._result_frame = result_frame

        controller.set_controlled_objects(
            self._dropdown_signals_frame,
            self._source_figure_frame,
            self._checkbox_types_signals_frame,
        )
        self._controller = controller

        self._reaction_mass = reaction_mass
        if limits is not None:
            print(limits)
            source_figure.add_infty_line(
                -abs(limits),
                "y",
                "bottom_limit",
                "limits",
                self._line_properties_limits,
            )
            source_figure.add_infty_line(
                abs(limits), "y", "upper_limit", "limits", self._line_properties_limits
            )

        self._checkbox_types_signals_frame.add_choise(
            "sweep",
            "types_signals",
            {"color": self._line_properties_sweep["line_color"]},
            self._controller,
        )

        self._checkbox_types_signals_frame.add_choise(
            "displacement",
            "types_signals",
            {"color": self._line_properties_displacement["line_color"]},
            self._controller,
        )

        self._checkbox_types_signals_frame.add_choise(
            "envelope",
            "types_signals",
            {"color": self._line_properties_envelope["line_color"]},
            self._controller,
        )

        self._checkbox_types_signals_frame.add_choise(
            "limits",
            None,
            {"color": self._line_properties_limits["line_color"]},
            self._controller,
        )

    def add_signal(self, sweep_signal: NamedSweep) -> None:
        plot = self._source_figure_frame.get_figure()

        plot.add_line(
            *sweep_signal.get_data(),
            str(sweep_signal),
            set(["sweep", str(sweep_signal)]),
            self._line_properties_sweep,
        )

        displacement = sweep_signal.integrate().integrate() / self._reaction_mass

        plot.add_line(
            *displacement.get_data(),
            f"{str(sweep_signal)}_displacement",
            set(["displacement", str(sweep_signal)]),
            self._line_properties_displacement,
        )

        plot.add_line(
            *sweep_signal.a_t.get_data(),
            f"{str(sweep_signal)}_a_t",
            set(["envelope", str(sweep_signal)]),
            self._line_properties_envelope,
        )

        self._dropdown_signals_frame.add_choise(str(sweep_signal))

    def show(self):
        return self._result_frame.get_output()



from ..named_signals.defaults.methods import extract_input


class NamedConfig:
    """Configuration for named_signals.

    - - -

    **NAMING_BY_ASSIGNMENT_CREATE**: `bool`
    > If `True` then my_relation = `NamedRelation`(...) and `str`(my_relation) ==
    'my_relation', otherwise false then my_relation = `NamedRelation`(...) and
    `str`(my_relation) == 'R_1' using
    `sweep_design.named_signals.header_signals.defaults.names.make_default_relation_name`
    with parameter 1.

    > In both situations, if name param name is specified, then
    my_relation = `NamedRelation`(..., name='MyName') and `str`(my_relation) == 'MyName'.

    - - -

    **NAMING_BY_ASSIGNMENT_MATH_OPERATION**: `bool`
    > This veriable is similar to  **NAMING_BY_ASSIGNMENT**, but you can't send
    name to a math operation.

    > If **NAMING_BY_ASSIGNMENT_MATH_OPERATION** is `False` then
    > my_relation = `NamedRelation`(..., name = 'MyName')
    > sum_relation = my_relation + my_relation
    > `str`(sum_relation) == 'MyName + MyName'
    using default method
    `sweep_design.named_signals.header_signals.defaults.names.make_default_names_operations`

    > otherwise if `True` then result was `str`(sum_relation) == 'sum_relation'

    - - -

    **NAMING_BY_ASSIGNMENT_OTHER_OPERATION**: `bool`
    > This variable is similar above for other operations.

    - - -

    **extract_input**
    > The method to convert input data to tuple of `numpy.ndarray`
    **x** and **y**.
    > Default methods is from
    `sweep_design.named_signals.defaults.methods.extract_input`

    > It's a simple method, but you can override them yourself.

    > **input**:
    >> **x**: `Any`
       **y**: `Any`

    > **output**:
    >> `Typle`[
        **x**: `numpy.ndarray`,
        **y**: `numpy.ndarray`
        ]

    The above methods can be overridden with your own here, or you can import the
    class `NamedConfig` somewhere and override it there.
     (They must be written according to the rules corresponding to
     the input and output parameters)

    """

    NAMING_BY_ASSIGNMENT_CREATE = True
    NAMING_BY_ASSIGNMENT_MATH_OPERATION = False
    NAMING_BY_ASSIGNMENT_OTHER_OPERATION = False

    extract_input = staticmethod(extract_input)



from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
from scipy import integrate  # type: ignore
from scipy.signal import hilbert, spectrogram  # type: ignore

from ...config.config import Config
from ..defaults.base_structures import BadInputError, RelationProtocol
from ..math_relation import Relation
from ..math_signal import Signal, Spectrum

InterpolateTime = Callable[[np.ndarray], np.ndarray]
CallFtatMethod = Callable[[Spectrum], Tuple[np.ndarray, np.ndarray, np.ndarray]]
Ftat = Union[np.ndarray, Callable[[np.ndarray], np.ndarray]]
Ftatr = Union[Relation, Ftat]

frequency = np.ndarray
time = np.ndarray
spectrogram_ = np.ndarray
envelope = np.ndarray

theta = np.ndarray


def simple_freq2time(spectrum: "Spectrum") -> Tuple[time, frequency, envelope]:
    """The simple method to extract frequency modulation from a prior spectrum.

    Properties:
    > spectrum: Spectrum

    Returns:
    > Tuple [
    >> time: np.ndarray,
    >> frequency: np.ndarray,
    >> amplitude_modulation: np.ndarray
    >]

    The amplitude modulation is constant.
    """
    amplitude_spectrum = spectrum.get_amp_spectrum()
    frequency, amplitude = amplitude_spectrum.get_data()
    n_spec = amplitude**2
    time = np.append(
        [0.0], ((n_spec[1:] + n_spec[:-1]) / (frequency[1:] - frequency[:-1])).cumsum()
    )
    coef = amplitude_spectrum.get_norm()
    amplitude_modulation = coef * np.ones(len(time))
    return time, frequency, amplitude_modulation


def pre_interpolate_time(
    x: Union[RelationProtocol, np.ndarray], y: np.ndarray = None
) -> InterpolateTime:
    """Thr interpolation of frequency and amplitude modulation functions.

    The method returns the interpolation function for frequency modulation
    and amplitude envelope. Since the sampling step is not the same,
    it is required to bring the desired sweep signal to the time axis.
    """
    if isinstance(x, Relation):
        x, y = x.get_data()

    def interpolate_time(time):
        nT = x * (time[-1] / x[-1])
        y2 = Config.interpolate_extrapolate_method(nT, y)
        return y2(time)

    return interpolate_time


# conversion adaptive sweep spectra FROM Ampl(freq) TO Freq(time)


def convert_freq2time(
    spectrum: "Spectrum", convert_method: CallFtatMethod
) -> Tuple[InterpolateTime, InterpolateTime]:
    """The method consists in obtaining the functions of frequency and amplitude modulation."""
    nT, f, a_t = convert_method(spectrum)
    # print(a_t)
    return pre_interpolate_time(nT, f), pre_interpolate_time(nT, a_t)


def get_info_from_aprior_data(
    t: Any, aprior_data: Any, f_a_t_method: CallFtatMethod
) -> Tuple[InterpolateTime, InterpolateTime, "Signal"]:
    """Get the frequency and amplitude modulation function from a prior data."""
    if isinstance(t, Spectrum):
        aprior_signal = t.get_signal()
        aprior_spectrum = t
    elif isinstance(t, Relation):
        aprior_signal = Signal(t)
        aprior_spectrum = aprior_signal.get_spectrum()
    elif isinstance(aprior_data, Spectrum):
        aprior_signal = aprior_data.get_signal()
        aprior_spectrum = aprior_data
    elif isinstance(aprior_data, Signal):
        aprior_signal = aprior_data
        aprior_spectrum = aprior_data.get_spectrum()
    elif isinstance(aprior_data, Relation):
        aprior_spectrum = Spectrum(aprior_data)
        aprior_signal = aprior_spectrum.get_signal()
    else:
        aprior_signal = Signal(t, aprior_data)
        aprior_spectrum = aprior_signal.get_spectrum()

    f_t, a_t = convert_freq2time(aprior_spectrum, f_a_t_method)

    return f_t, a_t, aprior_signal


def _extract_x_t(
    t: Optional[np.ndarray], x_t: Ftatr
) -> Tuple[Optional[np.ndarray], InterpolateTime]:

    if not callable(x_t):
        if isinstance(x_t, Relation):
            t, _ = x_t.get_data()
            b_x_t = pre_interpolate_time(x_t)
        else:
            if isinstance(t, np.ndarray) and isinstance(x_t, np.ndarray):
                if t.size != x_t.size:
                    calc_t = np.linspace(t[0], t[-1], x_t.size)
                    x_t = Config.interpolate_extrapolate_method(calc_t, x_t)
                    x_t = x_t(t)
                b_x_t = pre_interpolate_time(t, x_t)
            else:
                raise BadInputError("Not enough data: t or x_t")
    else:
        b_x_t = x_t
    return t, b_x_t


def get_info_from_ftat(
    t: Optional[np.ndarray], f_t: Optional[Ftatr], a_t: Optional[Ftatr]
) -> Tuple[Optional[np.ndarray], InterpolateTime, InterpolateTime]:
    """ """
    if f_t is None:

        def linear_time(time: np.ndarray) -> np.ndarray:
            return time

        f_t = linear_time

    t1, f_t = _extract_x_t(t, f_t)

    if a_t is None:

        def const_one(time: np.ndarray) -> np.ndarray:
            return np.ones(len(time))

        a_t = const_one

    t2, a_t = _extract_x_t(t, a_t)

    if t is None:
        if t1 is None and t2 is not None:
            t = t2
        elif t2 is None and t1 is not None:
            t = t1
        elif t2 is not None and t1 is not None:
            t = Config.get_common_x(t1, t2)

    return t, f_t, a_t


# ==============================================================================


def get_spectrogram(
    time: np.ndarray, amplitude: np.ndarray, dt: float = None
) -> Tuple[time, frequency, spectrogram_]:
    """Function to get spectrogram of the sweep signal.

    Using the scipy.signal.spectrogram function.
    """
    if dt is None:
        dt = time[1] - time[0]

    frequency, spectrogram_time, spectrogram_ = spectrogram(amplitude, 1 / (dt))

    return spectrogram_time, frequency, spectrogram_


def get_f_t(time: np.ndarray, amplitude: np.ndarray) -> Relation:
    """Get time-frequency function from sweep signal using the Hilbert transformation.

    Using the scipy.signal.hilbert function.
    """
    analytical_signal = hilbert(amplitude)
    result = np.append(
        [0.0],
        np.diff(np.unwrap(np.angle(analytical_signal)))
        / (2.0 * np.pi)
        / (time[1] - time[0]),
    )
    return Relation(time, result)


def get_a_t(time: np.ndarray, amplitude: np.ndarray) -> Relation:
    """Get envelop from sweep signal using the Hilbert transformation.

    Using the scipy.signal.hilbert function.
    """
    analytical_signal = hilbert(amplitude)
    return Relation(time, np.abs(analytical_signal))


def integrate_quad(
    f_t_function: Callable[[time], frequency], time: np.ndarray
) -> theta:
    """Integration.

    Integration of time-frequency function, using scipy.integrate.quad function.
    """
    return np.append(
        [0.0],
        np.array(
            [2 * np.pi * integrate.quad(f_t_function, time[0], t)[0] for t in time[1:]]
        ),
    )



from typing import Any, Union, Callable, Optional

from sweep_design.config.named_config import NamedConfig

from .named_relation import NamedRelation

from .header_signals.base_header import HeaderBase
from .header_signals.sweep_header import HeaderSpectrogram, HeaderSweep
from ..math_signals.defaults.base_structures import Spectrogram
from ..math_signals.math_signal import Spectrum
from ..math_signals.math_sweep import Sweep
from .named_signal import NamedSignal, NamedSpectrum


InName = Union[HeaderBase, str, Callable[[], str]]


class NamedSweep(NamedSignal):
    def __init__(
        self,
        time: Any,
        amplitude: Any = None,
        name: InName = None,
        category: str = "sweep",
        f_t: Any = None,
        a_t: Any = None,
        aprior_signal: Any = None,
    ) -> None:

        self.header: HeaderSweep = HeaderSweep(name, category)

        self._f_t = (
            super()._convert_input(f_t, self.header) if f_t is not None else None
        )
        self._a_t = (
            super()._convert_input(a_t, self.header) if a_t is not None else None
        )

        self.aprior_signal: Optional[NamedSignal] = (
            aprior_signal if aprior_signal is None else NamedSignal(aprior_signal)
        )

        if isinstance(aprior_signal, Spectrum):
            self.aprior_signal = NamedSignal(
                aprior_signal.get_signal(), name=self.header
            )
        elif isinstance(aprior_signal, NamedSpectrum):
            self.aprior_signal = aprior_signal.get_signal()
        elif isinstance(aprior_signal, NamedRelation):
            self.aprior_signal = NamedSignal(aprior_signal, name=aprior_signal.header)
        elif aprior_signal is not None:
            self.aprior_signal = NamedSignal(aprior_signal, name=self.header)
        else:
            self.aprior_signal = None

        r_f_t = None if self._f_t is None else self._f_t.relation
        r_a_t = None if self._a_t is None else self._a_t.relation
        r_aprior_signal = (
            None if self.aprior_signal is None else self.aprior_signal.relation
        )

        time, amplitude = NamedConfig.extract_input(time, amplitude)
        self._relation = Sweep(time, amplitude, r_f_t, r_a_t, None)

        self._spectrogram: Optional["NamedSpecrogram"] = None
        self.spectrum: Optional[NamedSpectrum] = None

    @property
    def f_t(self) -> NamedRelation:
        if self._f_t is None:
            header = self.header.get_f_t()
            self._f_t = NamedRelation(self.relation.frequency_time, name=header)
        return self._f_t

    @property
    def a_t(self) -> NamedRelation:
        if self._a_t is None:
            header = self.header.get_a_t()
            self._a_t = NamedRelation(self.relation.amplitude_time, name=header)
        return self._a_t

    @property
    def spectrogram(self) -> "NamedSpecrogram":
        if self._spectrogram is None:
            header = self.header.get_spectrogram()
            self._spectrogram = NamedSpecrogram(self._relation.spectrogram, name=header)
        return self._spectrogram

    def set_f_t(self, name: InName = None, category: str = None) -> None:

        header = self.header.get_f_t(name, category)
        self._f_t = NamedRelation(self.relation.frequency_time, name=header)

    def set_a_t(self, name: InName = None, category: str = None) -> None:
        header = self.header.get_a_t(name, category)
        self._a_t = NamedRelation(self.relation.amplitude_time, name=header)

    def set_spectrogram(self, name: InName = None, category: str = None) -> None:
        header = self.header.get_spectrogram(name, category)
        self._spectrogram = NamedSpecrogram(self.relation.spectrogram, name=header)


class NamedSpecrogram:
    def __init__(
        self, spectrogram: Spectrogram, name: InName = None, category: str = None
    ) -> None:
        self.header = HeaderSpectrogram(name, category)
        self.spectrogram = spectrogram

    @property
    def time(self):
        return self.spectrogram.time

    @property
    def frequency(self):
        return self.spectrogram.frequency

    @property
    def spectrogram_matrix(self):
        return self.spectrogram.spectrogram_matrix

    @property
    def category(self):
        return self.header.category



"""Prepared sweeps."""

from .dwell_sweep import get_dwell_sweep as get_dwell_sweep
from .linear_sweep import get_linear_sweep as get_linear_sweep
from .pseudorandom_shuffle import get_shuffle as get_shuffle
from .code_zinger import get_code_zinger as get_code_zinger
from .m_sequence import get_m_sequence as get_m_sequence
from .m_sequence import get_pure_m_sequence as get_pure_m_sequence



from typing import NamedTuple, Protocol, Tuple, runtime_checkable
import numpy as np
from enum import Enum


class NotEqualError(Exception):
    def __init__(self, *args: object) -> None:
        message = f"Different size x_array({args[0]}) and y_array({args[1]})"
        super().__init__(message)


class BadInputError(Exception):
    pass


class ConvertingError(Exception):
    def __init__(self, *args: object) -> None:
        message = f"Can not convert type {args[0]} into type {args[1]}."
        super().__init__(message)


class TypeFuncError(Exception):
    def __init__(self, *args: object) -> None:
        message = f'operation "{args[0]}" did not complite with type {args[1]} and type {args[2]}'
        super().__init__(message)


@runtime_checkable
class RelationProtocol(Protocol):
    @property
    def x(self) -> np.ndarray:
        ...

    @property
    def y(self) -> np.ndarray:
        ...

    def get_data(self) -> Tuple[np.ndarray, np.ndarray]:
        ...


class MathOperation(Enum):
    ADD = "__add__"
    RADD = "__radd__"
    SUB = "__sub__"
    RSUB = "__rsub__"
    MUL = "__mul__"
    RMUL = "__rmul__"
    TRUEDIV = "__truediv__"
    RTRUEDIV = "__rtruediv__"
    POW = "__pow__"
    RPOW = "__rpow__"


class BaseXY(NamedTuple):
    x: np.ndarray
    y: np.ndarray


class Spectrogram(NamedTuple):
    time: np.ndarray
    frequency: np.ndarray
    spectrogram_matrix: np.ndarray



import unittest

from ..math_relation import Relation
from ..math_signal import Signal, Spectrum
from .test_relation import PreTestRelation


class PreTestSignal:
    @staticmethod
    def pre_test_get_specrum(test_case: unittest.TestCase, signal_class: Signal):
        s = signal_class([0, 1, 2, 3, 4, 5], [0, 10, 20, 30, 40, 50])  # type: Signal
        sp = s.get_spectrum()
        test_case.assertIsInstance(sp, Spectrum)

    @staticmethod
    def pre_test_get_reverse(test_case: unittest.TestCase, signal_class: Signal):
        s = signal_class([0, 1, 2, 3, 4, 5], [0, 10, 20, 30, 40, 50])  # type: Signal

        rs = s.get_reverse_signal()
        test_case.assertIsInstance(rs, signal_class)

    @staticmethod
    def pre_test_math(test_case: unittest.TestCase, signal_class: Signal):
        s = signal_class([0, 1, 2, 3, 4, 5], [1, 2, 4, 6, 8, 10])  # type: Signal
        sp = s.get_spectrum()

        operation = ["__add__", "__sub__", "__mul__", "__truediv__", "__pow__"]
        for k in [s, sp]:
            for k2 in [s, sp]:
                for m in operation:
                    with test_case.subTest(k=k, k2=k2, m=m):
                        if isinstance(k2, Spectrum) and m in ["__truediv__"]:
                            continue
                        result = k.__getattribute__(m)(k2)
                        test_case.assertIsInstance(result, type(k))


class PreTestSpectrum:
    @staticmethod
    def pre_test_spectrum(test_case: unittest.TestCase, spectrum_class: Spectrum):

        sp = spectrum_class([0, 1, 2, 3, 4, 5], [1, 2, 4, 6, 8, 10])  # type: Spectrum

        amp = sp.get_amp_spectrum()
        test_case.assertIsInstance(amp, Relation)
        phase = sp.get_phase_spectrum()
        test_case.assertIsInstance(phase, Relation)
        s = sp.get_signal()
        test_case.assertIsInstance(s, Signal)
        rsp = sp.get_reverse_filter()
        test_case.assertIsInstance(rsp, spectrum_class)

        r1 = Relation([0, 1, 2, 3, 4, 5], [1, 2, 4, 6, 8, 10])
        r2 = Relation([0, 1, 2, 3, 4, 5], [1, 2, 4, 6, 8, 10])

        new_sp = spectrum_class.get_spectrum_from_amp_phase(r1, r2)
        test_case.assertIsInstance(new_sp, spectrum_class)


class TestSignalSpectrum(unittest.TestCase):
    pre_relation = PreTestRelation
    pre_signal = PreTestSignal
    pre_spectrum = PreTestSpectrum

    def test_input(self):
        self.pre_relation.pre_test_input(self, Signal)
        self.pre_relation.pre_test_input(self, Spectrum)

    def test_math(self):
        self.pre_relation.pre_test_math(self, Signal)
        self.pre_relation.pre_test_math(self, Spectrum)
        self.pre_signal.pre_test_math(self, Signal)

    def test_integrate_diff(self):
        self.pre_relation.pre_test_integrate_diff(self, Signal)
        self.pre_relation.pre_test_integrate_diff(self, Spectrum)

    def test_interpolate_extrapolate(self):
        self.pre_relation.pre_test_interpolate_extrapolate(self, Signal)
        self.pre_relation.pre_test_interpolate_extrapolate(self, Spectrum)

    def test_convolve_correlate(self):
        s1 = Signal([1, 2, 3, 4, 5], [10, 20, 30, 40, 50])
        sp1 = s1.get_spectrum()
        s2 = Signal([1, 2, 3, 4, 5], [10, 20, 30, 40, 50])
        sp2 = s2.get_spectrum()

        l = [s1, sp1, s2, sp2]

        for k in l:
            for k2 in l:
                for m in [Signal, Spectrum]:
                    self.subTest(k=k, k2=k2, m=m)
                    self.pre_relation.pre_test_convolve_correlate(self, k, k2, m)

    def test_spectrum(self):
        self.pre_spectrum.pre_test_spectrum(self, Spectrum)



from pathlib import Path
from typing import Sequence, Union
from loguru import logger

from scipy.io import savemat

from ...math_signals import Relation
from ...named_signals import NamedRelation

from .default import get_path_file_name


def write_mat_file(
    relation: Union[Relation, NamedRelation],
    file_name: Union[Path, str] = None,
    header: Sequence[str] = None,
) -> None:
    """Save the data to mat file."""

    file_name = get_path_file_name(relation, file_name, suffix=".mat")

    x, y = relation.get_data()
    if header is None:
        header = "x", "y"

    data = {header[0]: x, header[1]: y}
    savemat(file_name, data)

    logger.info("The file is saved in the following path: {}", file_name)



"""Module view.

The module provides several functions to get simple `GUI` to visualize the 
results in `jupyter notebook`.

The `sweep_design.view.base_view.abc_common_framer_grapher` module contains 
abstract classes for creating GUI elements.

The `sweep_design.view.framemakers` module contains concret classes for 
creating GUIs.

For example, classes that create graphics windows use libraries such as 
`bokeh` (http://docs.bokeh.org/) or `matplotlib` (https://matplotlib.org/), 
classes that create controls use the `ipywidget` (https://ipywidgets.readthedocs.io) 
library.

The modules described above can be used to create your own GUI.

Also, the `sweep_design.view` module contains several ready-made solutions. 
There are only 4 of them so far and they use the `ipywidget`, `matplotlib` and 
`bokeh` libraries:

- - -

The `sweep_design.view.view_general` module provides functions for general 
visualization (`sweep_design.view.view_general.get_general_view_bokeh_ipywidget` 
and `sweep_design.view.view_general.get_general_view_matplotlib_ipywidget` 
(the names of functions indicate which libraries are used))

Functions accept:
> **k_width**: `float` = 1. - the horizontal ratio of the render model window 
size to the screen size.  
> **k_height**: `float` = 1. - vertical ratio of the render model window 
size to the screen size.

Return a `sweep_design.view.view_general.view.GeneralView` object 
that represents a single render window. But by calling the function several 
times, you can get several new instances of the 
`sweep_design.view.view_general.view.GeneralView` class, and they can be 
combined as needed.

Next, you can build graphs and images using the 
`sweep_design.view.view_general.view.GeneralView.add_line` 
and `sweep_design.view.view_general.view.GeneralView.add_image` 
methods in the required corresponding instances, and then render in the 
corresponding cell by calling the 
`sweep_design.view.view_general.view.GeneralView.show()` method or by 
selecting an attribute of the object 
`sweep_design.view.view_general.view.GeneralView.result_view` and calling 
the `get_output()` method.

An example can be seen in the `examples.general_view.ipynb` folder in the project 
https://github.com/Omnivanitate/sweep_design  

- - -

The `sweep_design.view.view_source` module contains functions for visualizing a 
sweep signal for a vibration source. Displays the change in force as well as 
a graph of displacement.

A new visualization object is created using the get_view_source_bokeh_ipywidget 
function, which is passed the parameters:
> **reaction_mass**: `float` = 1. - reaction mass of the source,  
> **limits**: `float` = `None` - limits limiting its fluctuations,  
> **k_width**: `float` = 1. - horizontal ratio of the visualization model window 
size to the screen size,  
> **k_height**: `float` = 1. - vertical ratio of render model window size to 
screen size

The function returns an instance of the 
`sweep_design.view.view_source.source_view_builder.ViewSource` class.

The newly created instance can be passed a signal (an instance of the 
`sweep_design.named_signals.named_relation.NamedRelation` class) using the 
`sweep_design.view.view_source.source_view_builder.ViewSource.add_signal` 
method. And by calling 
the `sweep_design.view.view_source.source_view_builder.ViewSource.show()` 
method, visualize the result in the cell.

An example can be seen in the `examples.source_view.ipynb` folder in the project 
https://github.com/Omnivanitate/sweep_design 

- - -

The  module `sweep_design.view.view_sweep` contains the 
`sweep_design.view.view_sweep.bokeh_view.get_view_sweep_bokeh_ipywidget` 
function. 

The function accepts:
> **k_width**: `float` = 1 - horizontal ratio of the visualization model window 
size to the screen size,  
> **k_height**: `float` = 1 - vertical ratio of render model window size to screen size

The function retruns an instanse of 
`sweep_design.view.view_sweep.common_view_builder.CommonVeiwSweepBuilder` class 
to visualize sweep signals (`sweep_design.named_signals.named_sweep.NamedSweep`) 
and other signals (`sweep_design.named_signals.named_signal.NamedSignal`).

The instance of the 
`sweep_design.view.view_sweep.common_view_builder.CommonVeiwSweepBuilder` class 
contains an 
`sweep_design.view.view_sweep.common_view_builder.CommonVeiwSweepBuilder.add` 
method that can take the `sweep_design.named_signals.named_sweep.NamedSweep` 
or `sweep_design.named_signals.named_signal.NamedSignal` instance.

You can plot the result in a cell using the 
`sweep_design.view.view_sweep.common_view_builder.CommonVeiwSweepBuilder.show()` 
method.

An example can be seen in the `examples.sweep_view.ipynb` folder in the project 
https://github.com/Omnivanitate/sweep_design

- - -

`sweep_design.view.view_pilot_rm_bp` module for visualization of 
*pilot signal*, *plate signal* and *reaction mass signal*.
The module contains a function 
(`sweep_design.view.view_pilot_rm_bp.bokeh_view.get_pilot_rm_bp_view_bokeh_ipywidget`) 
that takes two arguments:

> **k_width**: `float` = 1 - horizontal ratio of the visualization model window 
size to the screen size,  
> **k_height**: `float` = 1. - vertical ratio of render model window size to 
screen size

The function returns a 
`sweep_design.view.view_pilot_rm_bp.bokeh_view.PilotRMBPBokehView` 
instance for rendering.

To add signals (**pilot**: `sweep_design.named_signals.named_sweep.NamedSweep`, 
**reaction_mass**: `sweep_design.named_signals.named_sweep.NamedSweep`, 
**base_plate**: `sweep_design.named_signals.named_sweep.NamedSweep`) 
the class instance has an 
`sweep_design.view.view_pilot_rm_bp.bokeh_view.PilotRMBPBokehView.add_pilot_rm_bp` 
method that accepts them.

You can plot the result in a cell using the 
`sweep_design.view.view_pilot_rm_bp.bokeh_view.PilotRMBPBokehView.show()` 
method.

An example can be seen in the `examples.pilot_rm_bp_view.ipynb` folder in 
the project https://github.com/Omnivanitate/sweep_design

"""

from .view_general import (
    get_general_view_bokeh_ipywidget as get_general_view_bokeh_ipywidget,
)
from .view_general import (
    get_general_view_matplotlib_ipywidget as get_general_view_matplotlib_ipywidget,
)
from .view_pilot_rm_bp import (
    get_pilot_rm_bp_view_bokeh_ipywidget as get_pilot_rm_bp_view_bokeh_ipywidget,
)
from .view_source import (
    get_view_source_bokeh_ipywidget as get_view_source_bokeh_ipywidget,
)
from .view_sweep import get_view_sweep_bokeh_ipywidget as get_view_sweep_bokeh_ipywidget



import math
import random

import numpy as np

from ..math_sweep import Sweep
from ..math_uncalcsweep import UncalculatedSweep
from ..utility_functions.tukey import tukey_a_t


class SegmentSmallerDtError(Exception):
    pass


def get_shuffle(
    time: np.ndarray,
    f_start=1.0,
    f_end=101.0,
    length_time_segments=0.5,
    round_number_freqency: int = None,
    time_end=10.0,
    t_tapper=1.0,
) -> Sweep:

    """Create shuffle sweep signal.

    t_tapper in seconds is used to apply tukey function at the end of dwell sweep signal.
    """

    dt = time[1] - time[0]
    time_end = time[-1]

    if dt >= length_time_segments:
        raise SegmentSmallerDtError(
            f"Length of shuffle time segment \
            (length_time_segments - {length_time_segments}) should be grate \
             then sample rate (dt - {dt})"
        )

    n_segments = math.ceil(time_end / length_time_segments)

    if not round_number_freqency:
        f_segment = np.linspace(f_start, f_end, n_segments + 1)
    else:
        f_segment = np.round(
            np.linspace(f_start, f_end, n_segments + 1), round_number_freqency
        )

    random.shuffle(f_segment)

    x = np.linspace(0.0, length_time_segments, math.ceil(length_time_segments / dt) + 1)
    my_cos = np.cos(x * np.pi / length_time_segments - np.pi) / 2 + 1 / 2
    f_t = []
    for f1, f2 in zip(f_segment[:-1], f_segment[1:]):
        f_t.extend((my_cos * (f2 - f1) + f1)[1:])
    f_t = np.array(f_t)

    t_correct = time
    f_t_correct = f_t[:-2]
    a_t = tukey_a_t(t_correct, t_tapper)

    unsw = UncalculatedSweep(t_correct, f_t_correct, a_t)

    return unsw()



from pathlib import Path
from typing import Union

from .defaults import get_read_file_name

from ...math_signals import Relation


def read_txt_file(file_name: Union[Path, str], sep="\t") -> Relation:
    """Read data from txt file."""
    file_name = get_read_file_name(file_name)
    with file_name.open() as f:
        lines = [line.strip().split(sep) for line in f.readlines()]
        x = []
        y = []
        for line in lines:
            if len(line) > 1:
                if line[0].isalnum() and line[1].isalnum():
                    x.append(line[0])
                    y.append(line[1])

    return Relation(x, y)



import numpy as np
from scipy.signal import butter, filtfilt, max_len_seq

from ..math_relation import Relation
from ..math_signal import Signal
from ..utility_functions import get_IMFs_ceemdan, tukey_a_t
import math
import random


def filtering(seq: Signal, f: float, type_filter: str):

    b, a = butter(3, f * 2 * seq.dx, type_filter)
    x, y = seq.get_data()
    y2 = filtfilt(b, a, y)
    return Signal(x, y2)


def get_m_sequence(
    len_seq: float,
    dt: float,
    reaction_mass: float = 1,
    displacement_limit: float = 1,
    force_limit: float = 1000,
    t_tapper=1.0,
    f_start=None,
    f_end=None,
    is_full=False,
) -> np.ndarray:

    start_len_seq = math.ceil(math.log((len_seq - 1), 2))
    start_seq = np.array([random.randint(0, 1) for _ in range(start_len_seq)])

    m_seq = max_len_seq(nbits=start_seq.size, state=start_seq)[0]
    t = np.linspace(0, dt * (m_seq.size - 1), m_seq.size)
    signal = 2 * Signal(t, m_seq) - 1

    if f_start is not None:
        signal = filtering(signal, f_start, "highpass")

    if f_end is not None:
        signal = filtering(signal, f_end, "lowpass")

    displacement = signal.integrate().integrate()

    imfs = get_IMFs_ceemdan(displacement)

    result_signal = sum(imfs[:-1])
    y = result_signal.y * tukey_a_t(result_signal.x, t_tapper)
    result_signal = Signal(result_signal.x, y)

    max_displacement: float = np.max(np.abs(result_signal.y))

    result_signal = result_signal / max_displacement * displacement_limit

    result_signal = result_signal.diff().diff() * reaction_mass

    max_force = np.max(np.abs(result_signal.y))

    if max_force > force_limit:
        result_signal = result_signal / max_force * force_limit

    if is_full:
        return result_signal

    print((len_seq - 1) * dt)

    return result_signal.select_data(x_end=(len_seq - 1) * dt)


def get_pure_m_sequence(len_seq: float, dt: float) -> Relation:

    len_seq = math.ceil(math.log((len_seq - 1), 2))
    start_seq = np.array([random.randint(0, 1) for _ in range(len_seq)])

    m_seq = max_len_seq(nbits=start_seq.size, state=start_seq)[0]
    t = np.linspace(0, dt * (m_seq.size - 1), m_seq.size)
    return Signal(t, m_seq)



from ...base_view.default_screen_properties import get_screen_params

_result = get_screen_params()

_width, _height = _result["width"], _result["height"]


def get_figure_frame(k_width=1.0, k_height=1.0):

    figure_frame = {
        "width": f"{int(_width*k_width)}px",
        "hieght": f"{int(_height*k_height)}px",
        "border": "solid 1px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
    }
    return figure_frame


def get_checkbox_frame(description: str, no_data: str = "", k_width=1.0, k_height=1.0):
    checkbox_frame = {
        "width": f"{int(_width*k_width)}px",
        "hieght": f"{int(_height**k_height)}px",
        "border": "solid 1px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "name_no_data": no_data,
        "description": description,
    }
    return checkbox_frame


def get_dropdown_frame(description: str, no_data: str):
    dropdown_frame = {
        "description": description,
        "name_no_data": no_data,
        "border": "solid 0px black",
        "margins": "10px, 10px, 10px, 10px",
        "padding": "5px, 5px, 5px, 5px",
        "name_no_data": "No sweep signals!",
        "description": "Sweep signals:",
    }
    return dropdown_frame



import numpy as np


def f_t_linear_function(f_start, f_end, time):
    return lambda t: t * (f_end - f_start) / (time[-1] - time[0]) + f_start


def f_t_linear_array(time: np.ndarray, f_start=0.0, f_end=1.0):
    return time * (f_end - f_start) / (time[-1] - time[0]) + f_start



import inspect
from typing import Optional, Tuple, Type, TypeVar, Union, Callable

import numpy as np

from ...math_signals.math_relation import Relation
from ...math_signals.defaults.base_structures import TypeFuncError
from ...math_signals.defaults.base_structures import MathOperation
from .base_header import HeaderBase
from .defaults import names as dfn
from .defaults.methods import make_name, make_category
from ...config.named_config import NamedConfig

Num = Union[float, int, complex]
InName = Union[HeaderBase, str, Callable[[], str]]
NR = TypeVar("NR", bound="HeaderRelation")


class BadNameCategory(Exception):
    pass


def _get_name_by_trace(position: int) -> str:
    frames_callers = inspect.stack()
    for n, k in enumerate(frames_callers):
        if "IPython" in k.filename:
            break

        names = frames_callers[n].frame.f_code.co_names
        if len(names) > 2:
            name = names[position]
        else:
            name = names[-1]
    return name


def set_name_by_trace(name: Optional[str], is_set=False, position=-1):
    if name is None:
        if is_set:
            name = _get_name_by_trace(position)
    return name


class HeaderRelation(HeaderBase):

    _quantity = 1
    _make_default_name = dfn.make_default_relation_name

    def __init__(
        self, name: InName = None, category: Optional[str] = "relation", **kwargs
    ) -> None:

        if name is None:
            if not NamedConfig.NAMING_BY_ASSIGNMENT_CREATE:
                name = type(self)._make_default_name(HeaderRelation._quantity)
                type(self)._quantity += 1
            else:
                name = _get_name_by_trace(-1)
        if category is None:
            category = "relation"

        super().__init__(name, category)

    @property
    def category(self) -> str:
        return self._category

    @category.setter
    def category(self, value) -> None:
        self._category = value

    @property
    def name(self) -> Callable[..., str]:
        return self._name

    @name.setter
    def name(self, value: InName) -> None:
        if callable(value) and isinstance(value(), str):
            self._name = value
        elif isinstance(value, HeaderBase):
            self._name = value._name
        else:
            str(value)

            def call() -> str:
                return str(value)

            self._name = call

    def select_data(
        self,
        x_start: Num,
        x_end: Num,
        name: InName = None,
        category="relation",
        **kwargs,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION)
        name = make_name(name, dfn.make_default_name_select_data, self, x_start, x_end)
        category = make_category(self, category)
        return HeaderRelation(name, category, **kwargs)

    def exp(self, name: InName = None, category: str = None) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION)
        name = make_name(name, dfn.make_default_name_exp, self)
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def diff(self, name: InName = None, category: str = None) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION)
        name = make_name(name, dfn.make_default_diff_name, self)
        category = make_category(self, category, dfn.make_diff_category_name)
        return HeaderRelation(name, category)

    def integrate(self, name: InName = None, category: str = None) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION)
        name = make_name(name, dfn.make_default_integrate_name, self)
        category = make_category(self, category, dfn.make_integrate_category_name)
        return HeaderRelation(name, category)

    def interpolate_extrapolate(
        self,
        new_x: Union[np.ndarray, Relation],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":
        if isinstance(new_x, Relation):
            new_x, _ = new_x.get_data()
        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION)
        name = make_name(name, dfn.make_default_interpolate_name, self, new_x)
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def shift(
        self, x_shift: Num, name: InName = None, category: str = None
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION)
        name = make_name(name, dfn.make_default_shift_name, self, x_shift)
        category = make_category(self, category)
        return HeaderRelation(name, category)

    @staticmethod
    def equalize(
        r1: "HeaderRelation",
        r2: "HeaderRelation",
        new_x: Union[Relation, np.ndarray],
        name1: InName = None,
        name2: InName = None,
        category1: str = None,
        category2: str = None,
    ) -> Tuple[HeaderBase, HeaderBase]:
        name1 = set_name_by_trace(
            name1, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION, -2
        )
        name2 = set_name_by_trace(
            name2, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION, -1
        )
        r1 = r1.interpolate_extrapolate(new_x, name1, category1)
        r2 = r2.interpolate_extrapolate(new_x, name2, category2)
        return r1, r2

    @classmethod
    def correlate(
        cls: Type["HeaderRelation"],
        r1: "HeaderRelation",
        r2: "HeaderRelation",
        name: InName = None,
        category=None,
        **kwargs,
    ) -> "HeaderRelation":

        if isinstance(r1, HeaderRelation) and isinstance(r2, HeaderRelation):
            name = set_name_by_trace(
                name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION
            )
            name = make_name(name, dfn.make_default_name_correlation, r1, r2)
            category = make_category(r1, category)
            return cls(name, category)
        else:
            raise TypeFuncError("Correlation", type(r1), type(r2))

    @classmethod
    def convolve(
        cls: Type["HeaderRelation"],
        r1: "HeaderRelation",
        r2: "HeaderRelation",
        name: InName = None,
        category: str = None,
        **kwargs,
    ) -> "HeaderRelation":

        if isinstance(r1, HeaderRelation) and isinstance(r2, HeaderRelation):
            name = set_name_by_trace(
                name, NamedConfig.NAMING_BY_ASSIGNMENT_OTHER_OPERATION
            )
            name = make_name(name, dfn.make_default_name_convolution, r1, r2)
            category = make_category(r1, category)
            return cls(name, category)
        else:
            raise TypeFuncError("Convolution", type(r1), type(r2))

    def __str__(self) -> str:
        return self.name()

    def __repr__(self) -> str:
        return f"Name: {self.name()} category: {self.category}"

    def __add__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.ADD
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __radd__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.RADD
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __sub__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.SUB
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __rsub__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.RSUB
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __mul__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.MUL
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __rmul__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.RMUL
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __truediv__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.TRUEDIV
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __rtruediv__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.RTRUEDIV
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __pow__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.POW
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __rpow__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":

        name = set_name_by_trace(name, NamedConfig.NAMING_BY_ASSIGNMENT_MATH_OPERATION)
        name = make_name(
            name, dfn.make_default_names_operations, self, other, MathOperation.RPOW
        )
        category = make_category(self, category)
        return HeaderRelation(name, category)

    def __iadd__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":
        return self.__add__(other, name, category)

    def __isub__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":
        return self.__sub__(other, name, category)

    def __imul__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":
        return self.__mul__(other, name, category)

    def __idiv__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":
        return self.__truediv__(other, name, category)

    def __ipow__(
        self,
        other: Union["HeaderRelation", Num],
        name: InName = None,
        category: str = None,
    ) -> "HeaderRelation":
        return self.__pow__(other, name, category)



from abc import ABCMeta, abstractmethod, abstractstaticmethod
from typing import Any, Dict, Generic, Iterable, Optional, Set, TypeVar, Union

from ..abc_common_framer_grapher.controller import CommonController

Output = TypeVar("Output")
Figure = TypeVar("Figure")
Controller = TypeVar("Controller", bound=CommonController)
Checkbox = TypeVar("Checkbox")
Dropbown = TypeVar("Dropbown")


class CommonFrameMaker(Generic[Figure, Controller], metaclass=ABCMeta):
    @abstractstaticmethod
    def get_figure_frame(
        figure: Figure, frame_properties: Dict[str, Any]
    ) -> "CommonFigureFrame[Figure]":
        pass

    @abstractstaticmethod
    def get_dropdown_frame(
        frame_properties: Dict[str, Any], controller: Controller
    ) -> "CommonDropdownFrame":
        pass

    @abstractstaticmethod
    def get_checkbox_frame(
        frame_properties: Dict[str, Any], controller: Controller
    ) -> "CommonCheckBoxFrame":
        pass


class CommonFrame(Generic[Output, Controller], metaclass=ABCMeta):
    def __init__(self, output: Output = None, controller: Controller = None) -> None:
        self._output = output
        self._controller = controller

    @property
    def controller(self) -> Optional[Controller]:
        return self._controller

    @controller.setter
    def controller(self, controller: Controller) -> None:
        self._controller = controller

    def get_output(self) -> Output:
        return self._output

    @abstractmethod
    def h_add(self, other: "CommonFrame") -> "CommonFrame":
        pass

    @abstractmethod
    def v_add(self, other: "CommonFrame") -> "CommonFrame":
        pass


class CommonFigureFrame(CommonFrame, Generic[Figure]):
    def __init__(self, figure: Figure, output: Output) -> None:
        super().__init__(output)
        self._figure = figure

    def get_figure(self) -> Figure:
        return self._figure


class CommonCheckBoxFrame(CommonFrame, Generic[Checkbox]):
    def __init__(self, output: Checkbox, controller: Controller = None) -> None:
        super().__init__(output, controller)
        self._checkboxs: Dict[str, CommonCheckBox] = {}

    @abstractmethod
    def add_choise(
        self,
        choise: str,
        groups: Union[str, Iterable[str]],
        checkbox_properties: Dict[str, Any],
        controller: Controller,
    ) -> None:
        pass

    @abstractmethod
    def return_choise(self, choise: str) -> None:
        pass

    @abstractmethod
    def remove_choise(self, choise: str) -> None:
        pass

    def remove_choise_by_groups(self, groups: Set[str]) -> None:
        for name_checkbox, checkbox in self._checkboxs.items():
            if groups.issubset(checkbox.get_groups()):
                self.remove_choise(name_checkbox)

    def return_choise_by_groups(self, groups: Set[str]) -> None:
        for name_checkbox, checkbox in self._checkboxs.items():
            if groups.issubset(checkbox.get_groups()):
                self.return_choise(name_checkbox)

    def get_checkboxs(self) -> Dict[str, "CommonCheckBox"]:
        return self._checkboxs

    def set_choise_by_name(self, name: str, value: bool) -> None:
        self._checkboxs[name].set_select(value)

    def set_choise_by_groups(self, groups: Union[str, Set[str]], value: bool) -> None:
        for choise_box in self._checkboxs.values():
            if groups.issubset(choise_box.get_groups()):
                choise_box.set_select(value)


class CommonCheckBox(CommonFrame, Generic[Checkbox, Controller]):
    def __init__(
        self,
        checkbox: Checkbox,
        name: str,
        groups: Union[str, Iterable[str]],
        controller: Controller,
    ) -> None:

        super().__init__(checkbox, controller)
        self._name = name
        if isinstance(groups, str):
            groups = set([groups])
        elif groups is not None:
            self._groups = set(groups)
        else:
            self._groups = set()
        self._checkbox = checkbox

    def get_name(self) -> str:
        return self._name

    def get_groups(self) -> Set[str]:
        return self._groups

    @abstractmethod
    def set_select(self, value: bool) -> None:
        pass


class CommonDropdownFrame(CommonFrame, Generic[Dropbown]):
    def __init__(
        self, dropdown: Dropbown, no_data_name: str, controller: Controller = None
    ) -> None:
        super().__init__(dropdown, controller)
        self._no_data_name = no_data_name
        self._dropdown = dropdown

    @abstractmethod
    def add_choise(self, choise: str) -> None:
        pass

    @abstractmethod
    def remove_choise(self, choise: str) -> None:
        pass

    @abstractmethod
    def get_choise(self) -> str:
        pass



from typing import Any, Callable, Optional, Type, TypeVar, Union

from ..math_signals.math_relation import Relation
from ..math_signals.math_signal import Signal, Spectrum
from .header_signals.base_header import HeaderBase
from .header_signals.signal_header import HeaderSignal, HeaderSpectrum
from .named_relation import NamedRelation
from ..config.named_config import NamedConfig

Num = Union[float, int, complex]
CR = TypeVar("CR", bound="NamedRelation")
InName = Union[HeaderBase, str, Callable[[], str]]
InData = Union[NamedRelation, "NamedSignal", "NamedSpectrum"]
InDataN = Union[NamedRelation, Relation, HeaderBase, Num]
CSP = TypeVar("CSP", bound="NamedSpectrum")
CS = TypeVar("CS", bound="NamedSignal")


class NamedSpectrum(NamedRelation):
    def __init__(
        self, f: Any, s_a: Any = None, name: InName = None, category="spectrum"
    ) -> None:
        f, s_a = NamedConfig.extract_input(f, s_a)
        self._relation: Spectrum = Spectrum(f, s_a)
        self.header: HeaderSpectrum = HeaderSpectrum(name=name, category=category)
        self.signal: Optional[NamedSignal] = None

    def get_signal(
        self,
        recalculate=False,
        start_time: float = None,
        name: InName = None,
        category: str = None,
    ) -> "NamedSignal":

        if self.signal is None or recalculate:
            signal = self._relation.get_signal(
                recalculate=recalculate, start_time=start_time
            )
            header = self.header.get_signal(name, category)
            self.signal = NamedSignal(signal, name=header)
        return self.signal

    def get_amp_spectrum(
        self, name: InName = None, category: str = "amp_spectrum"
    ) -> NamedRelation:
        relation = self._relation.get_amp_spectrum()
        header = self.header.get_amp_spectrum(name, category)
        return NamedRelation(relation, name=header)

    def get_phase_spectrum(
        self, name: InName = None, category: str = "phase_spectrum"
    ) -> NamedRelation:
        relation = self._relation.get_phase_spectrum()
        header = self.header.get_phase_spectrum(name, category)
        return NamedRelation(relation, name=header)

    def get_reverse_filter(
        self: CSP,
        percent: Union[float, int] = 5,
        subtrack_phase=True,
        f_start: float = None,
        f_end: float = None,
        name: InName = None,
        category: str = None,
    ) -> CSP:
        header = self.header.get_reverse_filter(name, category)
        spectrum = self._relation.get_reverse_filter(
            percent=percent,
            subtrack_phase=subtrack_phase,
            f_start=f_start,
            f_end=f_end,
            name=header,
        )
        return type(self)(spectrum, name=header)

    def add_phase(
        self: CSP, other: Any, name: InName = None, category: str = None
    ) -> CSP:
        c_other = super()._convert_input(other)
        header = self.header.add_phase(c_other.header, name, category)
        spectrum = self._relation.add_phase(c_other.relation, name=header)
        return type(self)(spectrum, name=header)

    def sub_phase(
        self: CSP, other: Any, name: InName = None, category: str = None
    ) -> CSP:
        c_other = super()._convert_input(other)
        header = self.header.sub_phase(c_other.header, name, category)
        spectrum = self._relation.add_phase(c_other.relation, name=header)
        return type(self)(spectrum, name=header)

    @classmethod
    def get_spectrum_from_amp_phase(
        cls: Type[CSP],
        r1: Any,
        r2: Any,
        name: InName = None,
        category="spectrum",
    ) -> CSP:
        c_r1 = super()._convert_input(r1)
        c_r2 = super()._convert_input(r2)
        header = HeaderSpectrum.get_spectrum_from_amp_phase(
            c_r1.header, c_r2.header, name, category
        )
        spectrum = Spectrum.get_spectrum_from_amp_phase(c_r1.relation, c_r2.relation)
        return cls(spectrum, name=header)


class NamedSignal(NamedRelation):
    def __init__(
        self, t: Any, a: Any = None, name: InName = None, category="signal"
    ) -> None:
        t, a = NamedConfig.extract_input(t, a)
        self._relation: Signal = Signal(t, a)
        self.header: HeaderSignal = HeaderSignal(name, category)
        self.spectrum: Optional[NamedSpectrum] = None

    def get_amplitude_spectrum(
        self,
        recalculate=False,
        is_start_zero=False,
        name: InName = None,
        category: str = None,
    ) -> NamedRelation:

        amplitude_spectrum = self._relation.get_amplitude_spectrum(
            recalculate=recalculate, is_start_zero=is_start_zero
        )
        header = self.header.get_amlitude_spectrum(name, category)
        return NamedRelation(amplitude_spectrum, name=header)

    def get_phase_spectrum(
        self,
        recalculate=False,
        is_start_zero=False,
        name: InName = None,
        category: str = None,
    ) -> NamedRelation:

        phase_spectrum = self._relation.get_phase_spectrum(
            recalculate=recalculate, is_start_zero=is_start_zero
        )
        header = self.header.get_phase_spectrum(name, category)
        return NamedRelation(phase_spectrum, name=header)

    def get_spectrum(
        self,
        recalculate=False,
        is_start_zero=False,
        name: InName = None,
        category: str = None,
    ) -> NamedSpectrum:
        if self.spectrum is None or recalculate:
            spectrum = self._relation.get_spectrum(
                recalculate=recalculate, is_start_zero=is_start_zero
            )
            header = self.header.get_spectrum(name, category)
            self.spectrum = NamedSpectrum(spectrum, name=header)
        return self.spectrum

    def get_reverse_signal(
        self: CS,
        percent: Union[float, int] = 5,
        subtrack_phase: bool = True,
        f_start: float = None,
        f_end: float = None,
        name: InName = None,
        category: str = None,
    ) -> CS:
        header = self.header.get_reverse_signal(name, category)
        signal = self._relation.get_reverse_signal(
            percent=percent,
            subtrack_phase=subtrack_phase,
            f_start=f_start,
            f_end=f_end,
        )
        return type(self)(signal, name=header)

    def add_phase(
        self: CS, other: Any, name: InName = None, category: str = None
    ) -> CS:
        r_other = super()._convert_input(other)
        header = self.header.add_phase(r_other.header, name, category)
        signal = self._relation.add_phase(r_other.relation)
        return type(self)(signal, name=header)

    def sub_phase(
        self: CS, other: Any, name: InName = None, category: str = None
    ) -> CS:
        r_other = super()._convert_input(other)
        header = self.header.sub_phase(r_other.header, name, category)
        signal = self._relation.sub_phase(r_other.relation)
        return type(self)(signal, name=header)



from typing import Any, Union

import numpy as np
from numpy.typing import NDArray
from ..config.sweep_config import SweepConfig

from .math_relation import Relation
from .math_signal import Signal

from .defaults.base_structures import Spectrogram


class Sweep(Signal):
    """Class `Sweep`.

    A class for analyzing changes in a signal over time, inherited from
    the `Signal` class.

    For analysis, you can use not only the sweep signal, but also other
    signals for which the spectrogram needs to be considered.

    When creating an instance of the class, the spectrogram is calculated.
    The method used to calculate the spectrogram is defined in the
    `SweepConfig` class. You can override it with your own.

    If the frequency vs. time and amplitude vs. time functions have not
    been passed, they are also calculated, the `get_frequency_time`,
    `get_amplitude_time` methods defined in the `SweepConfig` class are used.

    Perform the same operations as for the inherited class.

    **Properties**:
    > **time**: `Union[RelationProtocol, NDArray]`
    The Relation class, or a class derived from the Relation class, or
    an array_like object containing numbers(real or complex).

    > **amplitude**: `NDArray` = `None`
    `None` or array_like object containing real or complex numbers.

    > **frequency_time**: `Relation` = `None`
    This parameter describes the change in frequency versus time of the
    transmitted signal.

    > **amplitude_time**: `Relation` = `None`
    This parameter describes the change in amplitude envelop of signal
    from the time of the transmitted signal.

    > **aprior_signal**: `Signal` = `None`
    The signal used to create the sweep signal.

    """

    def __init__(
        self,
        time: Union[Relation, NDArray],
        amplitude: NDArray = None,
        frequency_time: Relation = None,
        amplitude_time: Relation = None,
        aprior_signal: Signal = None,
    ) -> None:

        super().__init__(time, amplitude)

        self.frequency_time = (
            frequency_time
            if frequency_time is not None
            else SweepConfig.get_f_t(self.x, self.y)
        )
        self.amplitude_time = (
            amplitude_time
            if amplitude_time is not None
            else SweepConfig.get_a_t(self.x, self.y)
        )

        spectrogram = SweepConfig.spectrogram_method(self.x, self.y, self.dx)
        self.spectrogram = Spectrogram(
            time=spectrogram[0],
            frequency=spectrogram[1],
            spectrogram_matrix=spectrogram[2],
        )
        self.aprior_signal = aprior_signal



from pathlib import Path
from typing import Union

from ...config import Config


class FileNotFoundError(Exception):
    """Raise an exception if the file is not found."""

    pass


def get_read_file_name(file_name: Union[Path, str]):
    """Define the file name to be read."""
    if isinstance(file_name, str):
        file_name = Path(file_name)

    if not file_name.is_absolute():
        file_name = Config.DEFAULT_PATH / file_name

    if not file_name.is_file():
        raise FileNotFoundError(f"File ({file_name}) does not exist ")

    return file_name



import logging
from typing import Tuple, Type, TypeVar, Union

import numpy as np
from numpy.typing import NDArray

from ..config import Config
from .defaults.base_structures import (
    BadInputError,
    MathOperation,
    NotEqualError,
    RelationProtocol,
    TypeFuncError,
)

Num = Union[float, int, complex]
R = TypeVar("R", bound="Relation")
R2 = TypeVar("R2", bound="Relation")


class Relation:
    """A representation of dependency y from x (y = f(x))

    The class describe a dependency between two sequences x, y,
    represented by real or complex numbers. The length of sequences must be
    the same, and the sampling rate must not change throughout the entire
    sequence.

    **Properties**:
    > **x**: `Union[RelationProtocol, NDArray]`
    The Relation class, or a class derived from the Relation class, or
    an array_like object containing numbers(real or complex).

    > **y**: `NDArray` = None.
    None or array_like object containing real or complex numbers.

    > **dx**: `float` = None.
    Sample rate x-axis.

    For the instance of `Relation` class, define the basic mathematical operstions:
    *additon (+), subtraction(-), multiplication('*'), devision(/),
    expopnentiation ('*''*') and their unary representation (+=, -=, *=, /=).
    The result of the operation is a new instance of the Relation class.

    Determined correlation and convolution between two instances
    (methods: correlate and convolve).

    How those operations will be calculated determined by the methods described
    in the Config class. Methods can be overridden if necessary
    (sweep_design.math_signals.config).

    WARNING!!! When inheriting the Relation class, it is important to write correctly
    constructor. It must match the constructor of the Relation class.
    Because some methods return a type(self)(...). For example,
    addition method (def __add__(self: R, other: Union['Relation', Num])
     -> R). Or predefine these methods in the inherited class.

    """

    def __init__(
        self,
        x: Union[RelationProtocol, NDArray],
        y: NDArray = None,
        dx: float = None,
        **kwargs,
    ) -> None:

        self._math_operation = Config.math_operation
        self._interpolate_extrapolate_method = Config.interpolate_extrapolate_method
        self._integrate_one_method = Config.integrate_one_method
        self._integrate_method = Config.integrate_method
        self._differentiate_method = Config.differentiate_method

        self._dx = dx

        if isinstance(x, RelationProtocol):
            self._x, self._y = x.get_data()
            if y is not None:
                logging.warning(f'x is instance of {type(x)}, "y" was ignored')
        else:
            if y is None:
                raise BadInputError("y absent. Not enough data!")

            if Config.CONVERT2ARRAY:
                x, y = np.array(x), np.array(y)

            if x.size != y.size:
                raise NotEqualError(x.size, y.size)

            self._x, self._y = x, y

    @property
    def x(self) -> np.ndarray:
        return self._x.copy()

    @property
    def dx(self) -> float:
        if self._dx is not None:
            return self._dx
        diff_x = np.diff(self._x)
        values, counts = np.unique(diff_x, return_counts=True)

        num = values[np.argmax(counts)]
        if num < 1:
            self._dx = 1 / round(1 / num)
        else:
            self._dx = round(num)
        return self._dx

    @dx.setter
    def dx(self, value: float) -> None:
        self._dx = value

    @property
    def y(self) -> np.ndarray:
        return self._y.copy()

    def get_data(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return the data of the called object."""
        return self._x.copy(), self._y.copy()

    def max(self) -> Num:
        return self._y.max()

    def min(self) -> Num:
        return self._y.min()

    def get_norm(self) -> float:
        """Get signal rate.

        Calculated in terms of signal energy.
        """
        x, y = self._x, self._y
        return self._integrate_one_method(y**2, x) / (self.dx)

    def select_data(self: R, x_start: Num = None, x_end: Num = None, **kwargs) -> R:
        """Select data using x-axis."""

        x, y = self.get_data()

        if x_start is None:
            x_start = x[0]

        if x_end is None:
            x_end = x[-1]

        is_selected = np.logical_and(
            np.greater_equal(self._x, x_start), np.less_equal(self._x, x_end)
        )

        return type(self)(x[is_selected], y[is_selected], **kwargs)

    def exp(self: R, **kwargs) -> R:
        x, y = self.get_data()
        return type(self)(x, np.exp(y), **kwargs)

    def diff(self: R, **kwargs) -> R:
        x, y = self.get_data()
        result = self._differentiate_method(x, y, self.dx)
        return type(self)(*result, **kwargs)

    def integrate(self: R, **kwargs) -> R:
        x, y = self.get_data()
        result = self._integrate_method(x, y, self.dx)
        return type(self)(*result, **kwargs)

    def interpolate_extrapolate(self: R, new_x: np.ndarray, **kwargs) -> R:
        new_y = self._interpolate_extrapolate_method(self._x, self._y)(new_x)
        return type(self)(new_x, new_y, **kwargs)

    def shift(self: R, x_shift: Num = 0, **kwargs) -> R:
        x, y = self.get_data()
        return type(self)(x + x_shift, y, **kwargs)

    @staticmethod
    def equalize(r1: R, r2: R2) -> Tuple[R, R2]:
        x1, _ = r1.get_data()
        x2, _ = r2.get_data()
        comparation = x1 == x2
        if isinstance(comparation, np.ndarray):
            if all(comparation):
                return r1, r2
        x_new = Config.get_common_x(x1, x2, r1.dx, r2.dx)
        r1 = r1.interpolate_extrapolate(x_new)
        r2 = r2.interpolate_extrapolate(x_new)
        return r1, r2

    @classmethod
    def correlate(cls: Type[R], r1: "Relation", r2: "Relation", **kwargs) -> R:

        if isinstance(r1, Relation) and isinstance(r2, Relation):
            result = Config.correlate_method(cls, r1, r2)
            return cls(*result, **kwargs)
        else:
            raise TypeFuncError("Correlation", type(r1), type(r2))

    @classmethod
    def convolve(cls: Type[R], r1: "Relation", r2: "Relation", **kwargs) -> R:

        if isinstance(r1, Relation) and isinstance(r2, Relation):
            result = Config.convolve_method(cls, r1, r2)
            return cls(*result, **kwargs)
        else:
            raise TypeFuncError("Convolution", type(r1), type(r2))

    @staticmethod
    def _operation(
        a: "Relation", b: Union["Relation", Num], name_operation: MathOperation
    ) -> Tuple[np.ndarray, np.ndarray]:
        logging.debug(f"Type of a: {type(a)}")
        logging.debug(f"Type of b: {type(b)}")
        if isinstance(b, RelationProtocol):
            r1, r2 = Relation.equalize(a, b)
            x, y1 = r1.get_data()
            _, y2 = r2.get_data()
            return a._math_operation(x, y1, y2, name_operation)
        elif isinstance(b, (float, int, complex)):
            x, y = a.get_data()
            return a._math_operation(x, y, b, name_operation)
        else:
            raise TypeFuncError(name_operation.value.strip("_"), type(a), type(b))

    def __add__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.ADD), **kwargs)

    def __radd__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.RADD), **kwargs)

    def __sub__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.SUB), **kwargs)

    def __rsub__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.RSUB), **kwargs)

    def __mul__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.MUL), **kwargs)

    def __rmul__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.RMUL), **kwargs)

    def __truediv__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.TRUEDIV), **kwargs
        )

    def __rtruediv__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(
            *self._operation(self, other, MathOperation.RTRUEDIV), **kwargs
        )

    def __pow__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.POW), **kwargs)

    def __rpow__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return type(self)(*self._operation(self, other, MathOperation.RPOW), **kwargs)

    def __iadd__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return self.__add__(other, **kwargs)

    def __isub__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return self.__sub__(other, **kwargs)

    def __imul__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return self.__mul__(other, **kwargs)

    def __idiv__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return self.__truediv__(other, **kwargs)

    def __ipow__(self: R, other: Union["Relation", Num], **kwargs) -> R:
        return self.__pow__(other, **kwargs)

    def __len__(self) -> int:
        return self._x.size

    def __getitem__(self, item: Union[float, slice]):
        if isinstance(item, float):
            idx = (np.abs(self._x - item)).argmin()
            return self._x[idx], self._y[idx]
        if isinstance(item, slice):
            return self.select_data(item.start, item.stop)



"""The project is intended for designing sweep signals.

> The package is written in `Python` for creating and developing sweep signals 
of varying complexity. It contains not only tools for creation, but also 
tools for analysis, visualization and other purposes. It can be used both for 
educational and production purposes.  

> It is convenient to use `Jupyter Lab` or `Jupyter Notebook` to speed up the 
development of signals, to compare their parameters with other signals, and 
to visualize them. To visualize graphs, the project contains packages 
corresponding to these tasks. (In code editors or IDEs such as `vscode`, 
rendering of images, graphs may not be correct. It's better to use 
`Jupyter Lab` directly.)  

> The project is designed so that you can easily change the creation of sweep 
signals. For example, write your own methods describing how the frequency 
and amplitude will change from the time of the sweep signal.  

> The project was created to be able to produce various sweep signals: 
implemented and not implemented by a vibration source, from simple ones, like a 
linear sweep signal, to complex ones, like a pseudo-random sweep signal.  

- - -
## Install.

> You can install the library using `pip install sweep_design`, or you can 
clone or load it from GitHub, add path to the module and install requremnt 
packages using the `poetry install` or `pip install -r requirement.txt`, 
or coping pieces of code and create your own.

- - -
## Using the library.

The library consists of five sub-modules.

* `sweep_design.config` - contains the project configuration.
* `sweep_design.dataio` - contains simple functions for loading and saving 
data. For a difficult situation, write your own.
* `sweep_design.math_signals` - There is the core of project. Here is the basic 
concept of creating signals, specectra and sweep-signals realising. 
* `sweep_design.named_signals` - There is a representation of the created 
`math_signals` object. For simplicity, each represented object from 
`named_signals` has an additional field like name and category. This is 
convenient when you work with many signals, spectra and you don't to get 
confused.
* `sweep_design.view` - There is a sub-module for visualasing results. 
Contains a few `GUI` for `jupyter notebook`.

- - -
For convenient you can directly import a few objects from modules 
`sweep_design`, such as 
```python
from sweep_design import Relation
```
instead
```python
from sweep_design.math_signals.math_relation import Relation
```
But both variants are correct.
"""

from .math_signals import Relation as Relation
from .math_signals import Signal as Signal
from .math_signals import Spectrum as Spectrum
from .math_signals import Sweep as Sweep
from .math_signals import UncalculatedSweep as UncalculatedSweep
from .math_signals import ApriorUncalculatedSweep as ApriorUncalculatedSweep

from .named_signals import NamedRelation as NamedRelation
from .named_signals import NamedSignal as NamedSignal
from .named_signals import NamedSpectrum as NamedSpectrum
from .named_signals import NamedSweep as NamedSweep
from .named_signals import NamedUncalcSweep as NamedUncalcSweep
from .named_signals import NamedApriorUcalcSweep as NamedApriorUcalcSweep



from typing import Any, Callable, Tuple, Union

import numpy as np

from ..math_signals.math_signal import Spectrum
from ..math_signals.math_uncalcsweep import ApriorUncalculatedSweep, UncalculatedSweep
from .header_signals.base_header import HeaderBase
from .header_signals.uncalcsweep_header import NamedApriorUcalcSweep, NamedUncalcSweep
from .named_sweep import NamedSweep

InName = Union[HeaderBase, str, Callable[[], str]]

CallFtatMethod = Callable[[Spectrum], Tuple[np.ndarray, np.ndarray, np.ndarray]]


class ComposedUncalcSweep:
    def __init__(
        self,
        t: Any = None,
        f_t: Any = None,
        a_t: Any = None,
        name: InName = None,
        category: str = None,
    ) -> None:

        self.uncalcsweep = UncalculatedSweep(t, f_t, a_t)
        self.header = NamedUncalcSweep(name, category)

    def __call__(
        self, time: Any = None, tht0=0.0, name: InName = None, category: str = None
    ) -> NamedSweep:
        sweep = self.uncalcsweep(time, tht0)
        header = self.header(sweep._x, name, category)
        return NamedSweep(sweep, name=header)


class ComposedApriorUncalcSweep(ComposedUncalcSweep):
    def __init__(
        self,
        t: Any = None,
        aprior_data: Any = None,
        ftat_method: CallFtatMethod = None,
        name: InName = None,
        category: str = None,
    ) -> None:

        self.uncalcsweep = ApriorUncalculatedSweep(t, aprior_data, ftat_method)
        self.header = NamedApriorUcalcSweep(name, category)



from typing import Union, Callable, Any

import numpy as np

from .base_header import HeaderBase
from .sweep_header import HeaderSweep
from .defaults import names as dfn

InName = Union[HeaderBase, str, Callable[[], str]]


class NamedUncalcSweep(HeaderBase):

    _quantity = 1

    def __init__(self, name: InName = None, category: str = None) -> None:

        if name is None:
            name = dfn.default_name_uncalc_sweep(NamedUncalcSweep._quantity)
            NamedUncalcSweep._quantity += 1

        if category is None:
            category = "c_sweep"

        super().__init__(name, category)

    @property
    def category(self) -> str:
        return self._category

    @category.setter
    def category(self, value) -> None:
        self._category = value

    @property
    def name(self) -> Callable[..., str]:
        return self._name

    @name.setter
    def name(self, value: InName) -> None:
        if callable(value) and isinstance(value(), str):
            self._name = value
        elif isinstance(value, HeaderBase):
            self._name = value._name
        else:
            str(value)

            def call() -> str:
                return str(value)

            self._name = call

    def __call__(
        self, time: np.ndarray, name: InName = None, category: str = None
    ) -> HeaderSweep:

        if name is None:
            dt = time[1] - time[0]
            T = time[-1] - time[0]
            name = dfn.default_name_call_uncalc_sweep(dt, T, self)

        if category is None:
            category = self.category

        return HeaderSweep(name=name, category=category)


class NamedApriorUcalcSweep(NamedUncalcSweep):

    _quantity = 1

    def __init__(self, name: InName = None, category: str = None) -> None:

        if name is None:
            name = dfn.default_name_aprior_uncalc_sweep(NamedApriorUcalcSweep._quantity)
            NamedApriorUcalcSweep._quantity += 1

        if category is None:
            category = "c_sweep"

        super().__init__(name, category)



from typing import Callable, Union, Optional
from abc import ABCMeta, abstractmethod


InName = Union["HeaderBase", str, Callable[..., str]]


class HeaderBase(metaclass=ABCMeta):
    def __init__(self, name: InName, category: str) -> None:

        if callable(name):
            self._name = name
            self._category = category
        elif isinstance(name, HeaderBase):
            self._name = name._name
            self._category = name._category
        else:
            r_name = str(name)

            def call() -> str:
                return r_name

            self._name = call
            self._category = category

    @property
    @abstractmethod
    def category(self) -> str:
        pass

    @property
    @abstractmethod
    def name(self) -> Callable[..., str]:
        pass
